# No shebang line, this module is meant to be imported
#
# Copyright 2013 Oliver Palmer
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import atexit
import logging
import json
import os
import re
import socket
import sys
import shutil
import tempfile
import time
from datetime import datetime
from functools import wraps, partial
from os import urandom
from os.path import basename, isfile
from random import randint, choice
from StringIO import StringIO
from textwrap import dedent
from urllib import urlopen

try:
    from httplib import OK, CREATED
except ImportError:  # pragma: no cover
    from http.client import OK, CREATED

from twisted.internet.base import DelayedCall
from twisted.trial.unittest import TestCase as _TestCase, SkipTest, FailTest

from pyfarm.core.config import read_env, read_env_bool
from pyfarm.core.enums import AgentState, PY26, STRING_TYPES
from pyfarm.agent.http.core.client import post
from pyfarm.agent.config import config, logger as config_logger
from pyfarm.agent.sysinfo import memory, cpu, system

try:
    from unittest.case import _AssertRaisesContext

except ImportError:  # copied from Python 2.7's source
    class _AssertRaisesContext(object):
        def __init__(self, expected, test_case, expected_regexp=None):
            self.expected = expected
            self.failureException = test_case.failureException
            self.expected_regexp = expected_regexp

        def __enter__(self):
            return self

        def __exit__(self, exc_type, exc_value, tb):
            if exc_type is None:
                try:
                    exc_name = self.expected.__name__
                except AttributeError:
                    exc_name = str(self.expected)
                raise self.failureException(
                    "{0} not raised".format(exc_name))
            if not issubclass(exc_type, self.expected):
                # let unexpected exceptions pass through
                return False
            self.exception = exc_value # store for later retrieval
            if self.expected_regexp is None:
                return True

            expected_regexp = self.expected_regexp
            if isinstance(expected_regexp, STRING_TYPES):
                expected_regexp = re.compile(expected_regexp)
            if not expected_regexp.search(str(exc_value)):
                raise self.failureException('"%s" does not match "%s"' %
                         (expected_regexp.pattern, str(exc_value)))
            return True

from voluptuous import Schema
from twisted.internet.defer import Deferred, succeed
from pyfarm.agent.entrypoints.parser import AgentArgumentParser
from pyfarm.agent.http.api.base import APIResource
from pyfarm.agent.http.core.template import DeferredTemplate
from pyfarm.agent.utility import dumps

ENABLE_LOGGING = read_env_bool("PYFARM_AGENT_TEST_LOGGING", False)
PYFARM_AGENT_MASTER = read_env("PYFARM_AGENT_TEST_MASTER", "127.0.0.1:80")

if ":" not in PYFARM_AGENT_MASTER:
    raise ValueError("$PYFARM_AGENT_TEST_MASTER's format should be `ip:port`")

os.environ["PYFARM_AGENT_TEST_RUNNING"] = str(os.getpid())


try:
    response = urlopen("http://" + PYFARM_AGENT_MASTER)
    PYFARM_MASTER_API_ONLINE = response.code == OK
except Exception as e:
    PYFARM_MASTER_API_ONLINE = False


class skipIf(object):
    """
    Wrapping a test with this class will allow the test to
    be skipped if ``should_skip`` evals as True.
    """
    def __init__(self, should_skip, reason):
        self.should_skip = should_skip
        self.reason = reason

    def __call__(self, func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if callable(self.should_skip) and self.should_skip() \
                    or self.should_skip:
                raise SkipTest(self.reason)
            return func(*args, **kwargs)
        return wrapper

def requires_master(function):
    """
    Any test decorated with this function will fail if the master could
    not be contacted or returned a response other than 200 OK for "/"
    """
    @wraps(function)
    def wrapper(*args, **kwargs):
        if not PYFARM_MASTER_API_ONLINE:
            raise FailTest("Could not connect to master")
        return function(*args, **kwargs)
    return wrapper


def create_jobtype(classname=None, sourcecode=None):
    """Creates a job type on the master and fires a deferred when finished"""
    if classname is None:
        classname = "Test%s" % urandom(8).encode("hex")

    if sourcecode is None:
        sourcecode = dedent("""
        from pyfarm.jobtypes.core.jobtype import JobType
        class %s(JobType):
            pass""" % classname)

    finished = Deferred()

    def posted(response):
        if response.code == CREATED:
            finished.callback(response.json())
        else:
            finished.errback(response.json())

    post(config["master_api"] + "/jobtypes/",
         callback=posted, errback=finished.errback,
         data={"name": classname,
               "classname": classname,
               "code": sourcecode})

    return finished


class FakeRequestHeaders(object):
    def __init__(self, test, headers):
        self.test = test
        self.test.assertIsInstance(headers, dict)

        for key, value in headers.items():
            headers[key.lower()] = value

        self.headers = headers

    def getRawHeaders(self, header):
        return self.headers.get(header)


class FakeRequest(object):
    def __init__(self, test, method, uri, headers=None, data=None):
        if headers is None:
            headers = {}

        if "Content-Type" not in headers:
            headers.update({"Content-Type": ["application/json"]})

        if data is not None:
            data = dumps(data)

        self.test = test
        self.method = method
        self.uri = uri
        self.code = None
        self.finished = None
        self.requestHeaders = FakeRequestHeaders(test, headers)
        self.content = StringIO()
        self._response = StringIO()

        if isinstance(data, STRING_TYPES):
            self.content.write(data)
            self.content.seek(0)

    def getHeader(self, header):
        return self.requestHeaders.getRawHeaders(header)

    def setResponseCode(self, code):
        self.test.assertIsNone(
            self.finished, "finished() called before setResponseCode()")
        self.code = code

    def write(self, data):
        self.test.assertIsNone(
            self.finished, "finished() called before write()")
        if not isinstance(data, STRING_TYPES):
            data = dumps(data)
        self._response.write(data)

    def finish(self):
        self.test.assertIsNone(self.finished, "finish() already called")
        self._response.seek(0)
        self.finished = True

    def response(self):
        self.test.assertIsNotNone(self.finished, "finish() not called")
        if not self._response.len:
            raise ValueError("Not content.")

        try:
            response = json.load(self._response)
        except ValueError:
            self._response.seek(0)
            response = self._response.read()

        self._response.seek(0)
        return response


class FakeAgent(object):
    def __init__(self, stopped=None):
        if stopped is None:
            stopped = Deferred()
        self.stopped = stopped

    def stop(self):
        if isinstance(self.stopped, Deferred):
            self.stopped.callback(None)
        return self.stopped


class ErrorCapturingParser(AgentArgumentParser):
    def __init__(self, *args, **kwargs):
        super(ErrorCapturingParser, self).__init__(*args, **kwargs)
        self.errors = []

    def error(self, message):
        self.errors.append(message)


class TestCase(_TestCase):
    POP_CONFIG_KEYS = []
    RAND_LENGTH = 8

    # Global timeout for all test cases.  If an individual test takes
    # longer than this amount of time to execute it will be stopped.  This
    # value should always be set to a value that's *much* longer than the
    # expected duration of the longest test case.
    timeout = 15

    # Override the default `assertRaises` which does not provide
    # context management.
    def assertRaises(self, excClass, callableObj=None, *args, **kwargs):
        if excClass is AssertionError and sys.flags.optimize:
            self.skipTest(
                "AssertionError will never be raised, running in optimized "
                "mode.")

        context = _AssertRaisesContext(excClass, self)
        if callableObj is None:
            return context
        with context:
            callableObj(*args, **kwargs)

    # Override the default `assertRaisesRegexp` which does not provide
    # context management.
    def assertRaisesRegexp(self, expected_exception, expected_regexp,
                           callable_obj=None, *args, **kwargs):
        if expected_exception is AssertionError and sys.flags.optimize:
            self.skipTest(
                "AssertionError will never be raised, running in optimized "
                "mode.")

        context = _AssertRaisesContext(
            expected_exception, self, expected_regexp)
        if callable_obj is None:
            return context
        with context:
            callable_obj(*args, **kwargs)

    def assertDateAlmostEqual(
            self, date1, date2,
            second_deviation=0, microsecond_deviation=1000000 / 2):
        self.assertIsInstance(date1, datetime)
        self.assertIsInstance(date2, datetime)
        self.assertEqual(date1.year, date2.year)
        self.assertEqual(date1.month, date2.month)
        self.assertEqual(date1.day, date2.day)
        self.assertEqual(date1.hour, date2.hour)
        self.assertEqual(date1.minute, date2.minute)
        self.assertEqual(date1.second, date2.second)
        self.assertApproximates(
            date1.second, date2.second, second_deviation)
        self.assertApproximates(
            date1.microsecond, date2.microsecond, microsecond_deviation)

    # back ports of some of Python 2.7's unittest features
    if PY26:
        def assertIsNone(self, obj, msg=None):
            if obj is not None:
                self.fail(self._formatMessage(msg, "%r is not None" % obj))

        def assertIsNotNone(self, obj, msg=None):
            if obj is None:
                self.fail(self._formatMessage(msg, "unexpectedly None"))

        def assertIsInstance(self, obj, cls, msg=None):
            if not isinstance(obj, cls):
                self.fail(
                    self._formatMessage(
                        msg, "%r is not an instance of %r" % (obj, cls)))

        def assertNotIsInstance(self, obj, cls, msg=None):
            if isinstance(obj, cls):
                self.fail(
                    self._formatMessage(
                        msg, "%r is an instance of %r" % (obj, cls)))

        def assertIn(self, containee, container, msg=None):
            if containee not in container:
                raise self.failureException(msg or "%r not in %r"
                                            % (containee, container))
            return containee

        def assertNotIn(self, containee, container, msg=None):
            if containee in container:
                raise self.failureException(msg or "%r in %r"
                                            % (containee, container))
            return containee

        def skipTest(self, reason):
            raise SkipTest(reason)

    def setUp(self):
        super(TestCase, self).setUp()

        try:
            self._pop_config_keys
        except AttributeError:
            self._pop_config_keys = []

        self._pop_config_keys.extend(self.POP_CONFIG_KEYS)
        self._pop_config_keys.extend([
            "agent",
            "jobs",
            "jobtypes",
            "restart_requested",
            "current_assignments",
            "last_master_contact"])

        DelayedCall.debug = True
        if not ENABLE_LOGGING:
            logging.getLogger("pf").setLevel(logging.CRITICAL)
        config_logger.disabled = 1
        self.prepare_config()
        config_logger.disabled = 0

    def prepare_config(self):
        for key in self._pop_config_keys:
            config.pop(key, None)

        config.update({
            "jobtypes": {},
            "agent-id": randint(1, 50000),
            "current_assignments": {},
            "agent_systemid": system.system_identifier(),
            "agent_http_retry_delay": 1,
            "agent_http_persistent_connections": False,
            "master": PYFARM_AGENT_MASTER,
            "agent_hostname": os.urandom(self.RAND_LENGTH).encode("hex"),
            "agent_ram": memory.total_ram(),
            "agent_cpus": cpu.total_cpus(),
            "agent_api_port": randint(10000, 50000),
            "free_ram": memory.free_ram(),
            "agent_time_offset": randint(-50, 50),
            "state": choice(AgentState),
            "start": time.time(),
            "agent_pretty_json": False,
            "agent_html_template_reload": True,
            "agent_master_reannounce": randint(5, 15)})

    def _rmdir(self, path, on_exit=True):
        try:
            shutil.rmtree(path)
        except (IOError, OSError):
            if on_exit:
                atexit.register(self._rmdir, path, on_exit=False)

    def create_test_file(self, content=None, dir=None, suffix=""):
        """
        Creates a test file on disk using :func:`tempfile.mkstemp`
        and uses the lower level file interfaces to manage it.  This
        is done to ensure we have more control of the file descriptor
        itself so on platforms such as Windows we don't have to worry
        about running out of file handles.
        """
        fd, path = tempfile.mkstemp(suffix=suffix, dir=dir, text=True)

        if content is not None:
            stream = os.fdopen(fd, "w")
            stream.write(content)
            stream.flush()
            os.fsync(stream.fileno())

            try:
                os.close(stream.fileno())
            except (IOError, OSError):
                pass
        else:
            try:
                os.close(fd)
            except (IOError, OSError):
                pass

        # self.addCleanup(self._closefd, fd)
        return path

    def create_test_directory(self, count=10):
        directory = tempfile.mkdtemp()
        self.addCleanup(self._rmdir, directory)

        files = []
        for _ in range(count):
            files.append(self.create_test_file(dir=directory))

        return directory, files


class BaseRequestTestCase(TestCase):
    HTTP_SCHEME = read_env("PYFARM_AGENT_TEST_HTTP_SCHEME", "http")
    DNS_HOSTNAME = config["agent_unittest"]["dns_test_hostname"]
    TEST_URL = config[
        "agent_unittest"]["client_api_test_url_%s" % HTTP_SCHEME]
    REDIRECT_TARGET = config["agent_unittest"]["client_redirect_target"]

    # DNS working?
    try:
        socket.gethostbyname(DNS_HOSTNAME)
    except socket.gaierror:
        RESOLVED_DNS_NAME = False
    else:
        RESOLVED_DNS_NAME = True

    # Basic http request working?
    try:
        urlopen(TEST_URL)
    except IOError:
        HTTP_REQUEST_SUCCESS = False
    else:
        HTTP_REQUEST_SUCCESS = True

    def setUp(self):
        if not self.RESOLVED_DNS_NAME:
            self.skipTest("Could not resolve hostname %s" % self.DNS_HOSTNAME)

        if not self.HTTP_REQUEST_SUCCESS:
            self.skipTest(
                "Failed to send an http request to %s" % self.TEST_URL)


class BaseHTTPTestCase(TestCase):
    URI = NotImplemented
    CLASS = NotImplemented
    CLASS_FACTORY = NotImplemented
    CONTENT_TYPES = NotImplemented

    # Only run the real _run if we're inside a child
    # class.
    def _run(self, methodName, result):
        if self.CLASS is NotImplemented:
            return succeed(True)

        if self.CLASS is not NotImplemented and self.URI is NotImplemented:
            self.fail("URI not set")

        self.assertIsInstance(self.CONTENT_TYPES, list, "CONTENT_TYPES not set")
        return super(BaseHTTPTestCase, self)._run(methodName, result)

    def setUp(self):
        super(BaseHTTPTestCase, self).setUp()
        self.agent = config["agent"] = FakeAgent()
        self.assertIsNotNone(self.CLASS, "CLASS not set")

    def instance_class(self):
        if self.CLASS_FACTORY is not NotImplemented:
            return self.CLASS_FACTORY()
        else:
            return self.CLASS()

    def test_instance(self):
        self.instance_class()

    def test_leaf(self):
        if self.URI.endswith("/"):
            self.assertTrue(self.CLASS.isLeaf)
        else:
            self.assertFalse(self.CLASS.isLeaf)

    def test_implements_methods(self):
        instance = self.instance_class()
        for method_name in instance.methods:
            if method_name == "head":
                continue

            self.assertTrue(
                hasattr(instance, method_name),
                "%s does not have method %s" % (self.CLASS, method_name))
            self.assertTrue(callable(getattr(instance, method_name)))

    def test_content_types(self):
        self.assertIsInstance(self.CLASS.CONTENT_TYPES, set)
        for content_type in self.CONTENT_TYPES:
            self.assertIn(content_type, self.CLASS.CONTENT_TYPES,
                          "missing content type %s" % content_type)

    def test_methods_exist_for_schema(self):
        self.assertIsInstance(self.CLASS.SCHEMAS, dict)
        instance = self.instance_class()
        methods = set(method.upper() for method in instance.methods)
        for method, schema in self.CLASS.SCHEMAS.items():
            self.assertIsInstance(schema, Schema)
            self.assertEqual(
                method.upper(), method,
                "method name in schema must be upper case")
            self.assertNotEqual(method, "GET", "cannot have schema for GET")
            self.assertIn(method, methods)

    def test_missing_schemas(self):
        missing_schemas = []
        for method in self.CLASS.LOAD_DATA_FOR_METHODS:
            if method not in self.CLASS.SCHEMAS:
                missing_schemas.append(method)

        if missing_schemas:
            self.skipTest(
                "WARNING: Missing schema(s) for %s"
                % ", ".join(missing_schemas))


class BaseAPITestCase(BaseHTTPTestCase):
    CONTENT_TYPES = ["application/json"]

    def setUp(self):
        super(BaseAPITestCase, self).setUp()
        self.assertIsNotNone(self.URI, "URI not set")
        self.get = partial(FakeRequest, self, "GET", self.URI)
        self.post = partial(FakeRequest, self, "POST", self.URI)
        self.put = partial(FakeRequest, self, "PUT", self.URI)

    def test_parent(self):
        self.assertIsInstance(self.instance_class(), APIResource)


class BaseHTMLTestCase(BaseHTTPTestCase):
    CONTENT_TYPES = ["text/html", "application/json"]

    def setUp(self):
        super(BaseHTMLTestCase, self).setUp()
        self.get = partial(FakeRequest, self, "GET")
        self.post = partial(FakeRequest, self, "POST")
        self.put = partial(FakeRequest, self, "PUT")

    def test_template_set(self):
        self.assertIsNot(self.CLASS.TEMPLATE, NotImplemented)

    def test_template_loaded(self):
        instance = self.instance_class()
        template = instance.template
        self.assertIsInstance(template, DeferredTemplate)
        self.assertEqual(basename(template.filename), self.CLASS.TEMPLATE)
        self.assertTrue(isfile(template.filename))



# No shebang line, this module is meant to be imported
#
# Copyright 2013 Oliver Palmer
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Memory
------

Provides information about memory including swap usage, system memory usage,
and general capacity information.
"""

import psutil
from pyfarm.core.utility import convert


def used_ram():
    """Amount of physical memory currently in use by applications"""
    return total_ram() - free_ram()


def free_ram():
    """Amount of physical memory free for application use"""
    return int(convert.bytetomb(psutil.virtual_memory().available))


def total_ram():
    """Total physical memory installed on the system"""
    return int(convert.bytetomb(psutil.virtual_memory().total))


def process_memory():
    """Total amount of memory in use by this process"""
    process = psutil.Process()
    return int(convert.bytetomb(process.memory_info().rss))


def total_consumption():
    """
    Total amount of memory consumed by this process and any
    child process spawned by the parent process.  This includes
    any grandchild processes.
    """
    parent = psutil.Process()
    total = parent.memory_info().rss

    for child_process in parent.children(recursive=True):
        try:
            total += child_process.memory_info().rss
        # Catch possible race condition
        except psutil.NoSuchProcess:
            pass

    return int(convert.bytetomb(total))



# No shebang line, this module is meant to be imported
#
# Copyright 2014 Oliver Palmer
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Config
++++++

Contains the endpoint for viewing and working with the configuration
on the agent.
"""

# NOTE: This module is named 'config' instead of 'configuration' like
# in the web ui to match the underlying object name and to differentiate
# the names a little more.

from pyfarm.agent.http.api.base import APIResource
from pyfarm.agent.config import config
from pyfarm.agent.utility import dumps, request_from_master


# TODO: add POST to modify keys
class Config(APIResource):
    isLeaf = False  # this is not really a collection of things

    def get(self, **kwargs):
        request = kwargs.get("request")

        if request is not None and request_from_master(request):
            config.master_contacted()

        return dumps(dict(config))



# No shebang line, this module is meant to be imported
#
# Copyright 2014 Oliver Palmer
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Job Type Internals
==================

Contains classes which contain internal methods for
the :class:`pyfarm.jobtypes.core.jobtype.JobType` class.
"""

import atexit
import imp
import os
import sys
import shutil
import tempfile
from collections import namedtuple
from errno import EEXIST, ENOENT
from datetime import datetime
from os.path import dirname, join, isfile, basename
from uuid import UUID
from functools import partial

try:
    import pwd
except ImportError:  # pragma: no cover
    pwd = NotImplemented

try:
    import grp
except ImportError:
    grp = NotImplemented

try:
    from httplib import (
        OK, INTERNAL_SERVER_ERROR, CREATED, ACCEPTED, CONFLICT)
except ImportError:  # pragma: no cover
    from http.client import (
        OK, INTERNAL_SERVER_ERROR, CREATED, ACCEPTED, CONFLICT)

from twisted.internet import reactor, threads
from twisted.internet.defer import Deferred, DeferredList, succeed

import treq

from pyfarm.core.enums import WINDOWS, INTEGER_TYPES, STRING_TYPES, WorkState
from pyfarm.agent.config import config
from pyfarm.agent.logger import getLogger
from pyfarm.agent.http.core.client import get, post, http_retry_delay
from pyfarm.jobtypes.core.log import STDOUT, STDERR, logpool
from pyfarm.jobtypes.core.process import ReplaceEnvironment, ProcessProtocol

USER_GROUP_TYPES = tuple(
    list(STRING_TYPES) + list(INTEGER_TYPES) + [type(None)])
ITERABLE_CONTAINERS = (list, tuple, set)

logcache = getLogger("jobtypes.cache")
logger = getLogger("jobtypes.core")
logfile = getLogger("jobtypes.log")
ProcessData = namedtuple(
    "ProcessData", ("protocol", "started", "stopped", "log_identifier"))


class Cache(object):
    """Internal methods for caching job types"""
    cache = {}
    JOBTYPE_VERSION_URL = \
        "%(master_api)s/jobtypes/%(name)s/versions/%(version)s"
    CACHE_DIRECTORY = config.get("jobtype_cache_directory", "")

    if not CACHE_DIRECTORY:  # pragma: no cover
        CACHE_DIRECTORY = None  # make sure it's None
        logger.warning("Job type cache directory has been disabled.")

    else:
        try:
            os.makedirs(CACHE_DIRECTORY)

        except OSError as e:  # pragma: no cover
            if e.errno != EEXIST:
                logger.error(
                    "Failed to create %r.  Job type caching is "
                    "now disabled.", CACHE_DIRECTORY)
                raise
        else:
            logger.info("Created job type cache directory %r", CACHE_DIRECTORY)

        logger.debug("Job type cache directory is %r", CACHE_DIRECTORY)

    @classmethod
    def _download_jobtype(cls, name, version):
        """
        Downloads the job type specified in ``assignment``.  This
        method will pass the response it receives to :meth:`_cache_jobtype`
        however failures will be retried.
        """
        logger.debug("Downloading job type %r version %s", name, version)
        url = str(cls.JOBTYPE_VERSION_URL % {
            "master_api": config["master_api"],
            "name": name, "version": version})

        result = Deferred()
        download = lambda *_: \
            get(url,
                callback=result.callback,
                errback=lambda: reactor.callLater(http_retry_delay(), download))
        download()
        return result

    @classmethod
    def _cache_filepath(cls, cache_key, classname, version):
        return str(join(
            cls.CACHE_DIRECTORY, "%s_%s_v%s.py" % (
                cache_key, classname, version)))

    @classmethod
    def _cache_key(cls, assignment):
        return assignment["jobtype"]["name"], assignment["jobtype"]["version"]

    @classmethod
    def _jobtype_download_complete(cls, response, cache_key):
        # Server is offline or experiencing issues right
        # now so we should retry the request.
        if response.code >= INTERNAL_SERVER_ERROR:
            logger.debug(
                "Could not download jobtype because of internal server error.")
            return reactor.callLater(
                http_retry_delay(),
                response.request.retry)

        downloaded_data = response.json()

        if not config["jobtype_enable_cache"]:
            logger.debug("Jobtype cache is disabled, loading the jobtype.")
            return cls._load_jobtype(downloaded_data, None)

        else:
            # When the download is complete, cache the results
            logger.debug("Caching the jobtype locally.")
            caching = cls._cache_jobtype(cache_key, downloaded_data)
            caching.addCallback(
                lambda result: cls._load_jobtype(*result))
            return caching


    @classmethod
    def _cache_jobtype(cls, cache_key, jobtype):
        """
        Once the job type is downloaded this classmethod is called
        to store it on disk.  In the rare even that we fail to write it
        to disk, we store it in memory instead.
        """
        if isinstance(cache_key, tuple):
            cache_key = cache_key[0]

        assert isinstance(cache_key, STRING_TYPES)
        assert isinstance(jobtype, dict)
        filename = cls._cache_filepath(
            cache_key, jobtype["classname"], jobtype["version"])
        success = Deferred()
        jobtype = jobtype.copy()

        def write_to_disk(filename):
            parent_dir = dirname(filename)
            try:
                os.makedirs(parent_dir)
            except (IOError, OSError) as e:  # pragma: no cover
                if e.errno != EEXIST:
                    logger.error("Failed to create %s: %s", parent_dir, e)
            else:
                logger.debug("Created %s", parent_dir)

            if isfile(filename):  # pragma: no cover
                logcache.warning("%s is already cached on disk", filename)
                jobtype.pop("code", None)
                return jobtype, filename

            try:
                with open(filename, "w") as stream:
                    stream.write(jobtype["code"])

            # If the above fails, use a temp file instead
            except (IOError, OSError) as e:  # pragma: no cover
                _, tmpfilepath = tempfile.mkstemp(suffix=".py")
                logcache.warning(
                    "Failed to write %s, using %s instead: %s",
                    filename, tmpfilepath, e)

                with open(tmpfilepath, "w") as stream:
                    stream.write(jobtype["code"])

                jobtype.pop("code", None)
                return jobtype, tmpfilepath

            else:
                logger.debug(
                    "Wrote job type %s version %s to %s",
                    jobtype["name"], jobtype["version"], filename)
                jobtype.pop("code", None)
                return jobtype, filename

        def written_to_disk(results):
            jobtype, filename = results
            cls.cache[cache_key] = (jobtype, filename)
            logcache.info("Created cache for %r at %s", cache_key, filename)
            success.callback((jobtype, filename))

        def failed_to_write_to_disk(error):  # pragma: no cover
            logcache.error(
                "Failed to write job type cache to disk, will use "
                "memory instead: %s", error)

            # The code exists in the job type because it's
            # only removed on success.
            cls.cache[cache_key] = (jobtype, None)
            success.callback((jobtype, None))

        # Defer the write process to a thread so we don't
        # block the reactor if the write is slow
        logger.debug(
            "Caching job type %s version %s to %s",
            jobtype["classname"], jobtype.get("version", "?"), filename)
        writer = threads.deferToThread(write_to_disk, filename)
        writer.addCallbacks(written_to_disk, failed_to_write_to_disk)
        return success

    @classmethod
    def _module_for_jobtype(cls, jobtype):
        return "pyfarm.jobtypes.cached.%s%s%s" % (
            jobtype["classname"], jobtype["name"], jobtype["version"])

    @classmethod
    def _load_jobtype(cls, jobtype, filepath):
        def load_jobtype(jobtype_data, path):
            module_name = cls._module_for_jobtype(jobtype_data)

            # Create or load the module
            if filepath is not None:
                logger.debug("Attempting to load module from file path %s",
                             filepath)
                try:
                    module = imp.load_source(module_name, path)
                except Exception as e:
                    type = sys.exc_info()[0]
                    value = sys.exc_info()[1]
                    logger.error("Importing module from jobtype file failed: "
                                 "%s, value: %s", type, value)
                    raise
            else:
                logcache.warning(
                    "Loading (%s, %s) directly from memory",
                    jobtype_data["name"], jobtype_data["version"])

                module = imp.new_module(module_name)
                exec jobtype_data["code"] in module.__dict__
                sys.modules[module_name] = module

            logger.debug("Returning class %s from module",
                         jobtype_data["classname"])
            return getattr(module, jobtype_data["classname"])

        # Load the job type itself in a thread so we limit disk I/O
        # and other blocking issues in the main thread.
        return threads.deferToThread(load_jobtype, jobtype, filepath)


class Process(object):
    """Methods related to process control and management"""
    logging = {}

    def __init__(self):
        self.start_called = False
        self.stop_called = False
        self._stopped_deferred = None
        self._start_deferred = None

    @property
    def stopped_deferred(self):
        if not self.start_called:
            raise RuntimeError("Not yet started")
        return self._stopped_deferred

    @property
    def start_deferred(self):
        if not self.start_called:
            raise RuntimeError("Not yet started")
        return self._stopped_deferred

    @stopped_deferred.setter
    def stopped_deferred(self, value):
        assert self.start_called
        assert self._stopped_deferred is None
        assert isinstance(value, Deferred)
        self._stopped_deferred = value

    @start_deferred.setter
    def start_deferred(self, value):
        assert self.start_called
        assert self._start_deferred is None
        assert isinstance(value, Deferred)
        self._start_deferred = value

    def _before_spawn_process(self, command, protocol):
        logger.debug(
            "%r._before_spawn_process(%r, %r)", self, command, protocol)
        self.before_spawn_process(command, protocol)

    def _spawn_twisted_process(
            self, _, command, process_protocol, kwargs):
        """
        Handles the spawning of the process itself using
        :func:`reactor.spawnProcess`.

        :param tuple _:
            An ignored argument containing the protocol id and
            csv log file
        """
        self._before_spawn_process(command, process_protocol)

        # The way Twisted handles the env keyword varies by platform.  To
        kwargs.setdefault("env", None)
        if kwargs["env"] is not None:
            raise RuntimeError(
                "The `env` keyword should always be set to None.")

        with ReplaceEnvironment(command.env):
            reactor.spawnProcess(process_protocol, command.command, **kwargs)

    def _start(self):
        """
        The internal method that gets called to start the job type.  Usually
        this calls :meth:`start` though more advanced behaviors could
        override this and :meth`start`.

        .. warning::

            Read the source code before overriding this method on your
            own.  This method sets a couple of instance variables and returns
            a tuple that is relied upon elsewhere.
        """
        # Make sure _start() is not called twice
        if self.start_called:
            raise RuntimeError("%s has already been started" % self)

        self._before_start()
        logger.debug("%r.start()", self.__class__.__name__)
        self.start()
        self.start_called = True
        logger.debug("Collecting started deferreds from spawned processes")

        if not self.processes:
            logger.warning(
                "There are no started deferreds present, using succeed([]) for "
                "self.started_deferred")
            self.started_deferred = succeed([])
        else:
            logger.debug("Making deferred list for self.started_deferred")
            self.started_deferred = DeferredList(
                [process.started for process in self.processes.values()])

        self.stopped_deferred = Deferred()
        return self.started_deferred, self.stopped_deferred

    def _stop(self):
        if self.stop_called:
            raise RuntimeError("%s has already been stopped" % self)

        return self.stop()

    def _before_start(self):
        logger.debug("%r._before_start()", self)
        self.before_start()

    def _process_started(self, protocol):
        """
        Called by :meth:`.ProcessProtocol.connectionMade` when a process has
        started running.
        """
        logger.debug("%r._process_started(%r)", self, protocol)
        logpool.log(protocol.uuid, STDOUT, "Started %r" % protocol)
        process_data = self.processes[protocol.uuid]
        process_data.started.callback(protocol)
        if not self.stop_called:
            self.process_started(protocol)
        else:
            self.stop()

    def _process_stopped(self, protocol, reason):
        """
        Internal implementation for :meth:`process_stopped`.

        If ``--capture-process-output`` was set when the agent was launched
        all standard output from the process will be sent to the stdout
        of the agent itself.  In all other cases we send the data to the
        logger pool so it can be stored in a file without blocking the
        event loop.
        """
        logger.info("%r stopped (code: %r)", protocol, reason.value.exitCode)
        process_data = self.processes.pop(protocol.uuid)

        if self.is_successful(reason):
            logpool.log(
                protocol.uuid, STDOUT,
                "Process has terminated successfully, code %s" %
                reason.value.exitCode)
        else:
            self.failed_processes.add((protocol, reason))
            logpool.log(
                protocol.uuid, STDOUT,
                "Process has not terminated successfully, code %s" %
                reason.value.exitCode)

        try:
            self.process_stopped(protocol, reason)
        except Exception as e:
            logger.error("Exception caught from process_stopped: %s", e)
        logpool.close_log(protocol.uuid)
        process_data.stopped.callback(reason)

        upload_deferred = self._upload_logfile(process_data.log_identifier)

        # If there are no processes running at this point, we assume
        # the assignment is finished
        if len(self.processes) == 0:
            if not self.failed_processes:
                logger.info("Processes in assignment %s stopped, no failures",
                            self)
            else:
                logger.warning("There was at least one failed process in the "
                               "assignment %s", self)
            self.stopped_deferred.callback(None)
        return upload_deferred

    def _spawn_process(self, command):
        """
        Starts one child process using input from :meth:`command_data`.
        Job types should never start child processes through any other
        means.  The only exception to this rule is code that resides in
        :meth:`prepare_for_job`, which should use
        :meth:`spawn_persistent_job_process` instead.

        :raises OSError:
            Raised if `working_dir` was provided but the provided
            path does not exist

        :raises EnvironmentError:
            Raised if an attempt is made to change the user or
            group without root access.  This error will only occur on
            Linux or Unix platforms.
        """
        process_protocol = self.PROCESS_PROTOCOL(self)

        if not isinstance(process_protocol, ProcessProtocol):
            raise TypeError("Expected ProcessProtocol for `protocol`")

        # The first argument should always be the command name by convention.
        # Under Windows, this needs to be the whole path, under POSIX only the
        # basename.
        if WINDOWS:
            arguments = [command.command] + list(command.arguments)
        else:
            arguments = [basename(command.command)] + list(command.arguments)

        # WARNING: `env` should always be None to ensure the same operation
        # of the environment setup across platforms.  See Twisted's
        # documentation for more information on why `env` should be None:
        #    http://twistedmatrix.com/documents/current/api/
        #    twisted.internet.interfaces.IReactorProcess.spawnProcess.html
        kwargs = {"args": arguments, "env": None}
        uid, gid = self.get_uid_gid(command.user, command.group)

        if uid is not None:
            kwargs.update(uid=uid)

        if gid is not None:
            kwargs.update(gid=gid)

        # Capture the protocol instance so we can keep track
        # of the process we're about to spawn and start the
        # logging thread.
        # TODO: return data from this function, we don't want to be working
        # with Deferred object in a public method
        log_path = self.get_csvlog_path(process_protocol.uuid)
        deferred = logpool.open_log(process_protocol, log_path)
        deferred.addCallback(
            self._spawn_twisted_process, command, process_protocol, kwargs)
        self.processes[process_protocol.uuid] = ProcessData(
            protocol=process_protocol, started=Deferred(), stopped=Deferred(),
            log_identifier=basename(log_path))

        self._register_logfile_on_master(basename(log_path))

        return deferred

    def _process_output(self, protocol, output, stream):
        """
        Called by :meth:`.ProcessProtocol.outReceived` and
        :meth:`.ProcessProtocol.errReceived` whenever output is produced
        by a process.  This method will wire up the proper calls under the
        hood to process the output.
        """
        if stream == STDOUT:
            line_fragments = self._stdout_line_fragments
            preprocessor = self.preprocess_stdout
            line_handler = self.handle_stdout_line

        elif stream == STDERR:
            line_fragments = self._stderr_line_fragments
            preprocessor = self.preprocess_stderr
            line_handler = self.handle_stderr_line

        else:
            raise ValueError("Expected STDOUT or STDERR for `stream`")

        # Proprocess before we call process_output
        new_output = preprocessor(protocol, output)
        if isinstance(new_output, STRING_TYPES):
            output = new_output

        self.process_output(protocol, output, line_fragments, line_handler)

    def _has_running_processes(self):
        """
        Internal functionto determine whether the batch represented by this
        instance still has running child processes.
        """
        for process in self.processes.values():
            if process.protocol.running():
                return True

        return False

    def _register_logfile_on_master(self, log_path):
        def post_logfile(task, log_path, delay=0):
            url = "%s/jobs/%s/tasks/%s/attempts/%s/logs/" % (
                config["master_api"], self.assignment["job"]["id"], task["id"],
                task["attempt"])
            data = {"identifier": log_path,
                    "agent_id": self.node()["id"]}
            post_func = partial(
                post, url, data=data,
                callback=lambda x: result_callback(task, log_path, x),
                errback=lambda x: error_callback(task, log_path, x))
            reactor.callLater(delay, post_func)

        def result_callback(task, log_path, response):
            if 500 <= response.code < 600:
                delay = http_retry_delay()
                logger.error(
                    "Server side error while registering log file %s for "
                    "task %s (frame %s) in job %s (id %s), status code: %s. "
                    "Retrying. in %s seconds",
                    log_path, task["id"], task["frame"],
                    self.assignment["job"]["title"],
                    self.assignment["job"]["id"], response.code, delay)
                post_logfile(task, log_path, delay=delay)

            # The server will return CONFLICT if we try to register a logfile
            # twice
            elif response.code not in [OK, CONFLICT, CREATED]:
                # Nothing else we could do about that, this is
                # a problem on our end.
                logger.error(
                    "Could not register logfile %s for task %s (frame %s) in "
                    "job %s (id %s), status code: %s. This is a client side "
                    "error, giving up.",
                    log_path, task["id"], task["frame"],
                    self.assignment["job"]["title"],
                    self.assignment["job"]["id"], response.code)

            else:
                logger.info("Registered logfile %s for task %s on master",
                            log_path, task["id"])

        def error_callback(task, log_path, failure_reason):
            delay = http_retry_delay()
            logger.error(
                "Error while registering logfile %s for task %s on master: "
                "\"%s\", retrying in %s seconds.",
                log_path, task["id"], failure_reason, delay)
            post_logfile(task, log_path, delay=delay)

        for task in self.assignment["tasks"]:
            post_logfile(task, log_path)

    def _upload_logfile(self, log_identifier):
        path = join(config["jobtype_task_logs"], log_identifier)
        url = "%s/jobs/%s/tasks/%s/attempts/%s/logs/%s/logfile" % (
                config["master_api"], self.assignment["job"]["id"],
                self.assignment["tasks"][0]["id"],
                self.assignment["tasks"][0]["attempt"],
                log_identifier)
        upload_deferred = Deferred()

        def upload(url, log_identifier, delay=0):
            logfile = open(path, "rb")
            if delay != 0:
                reactor.callLater(delay, upload, url,
                                  log_identifier=log_identifier)
            else:
                # FIXME persistent=False is a workaround to help with some
                # problems in unit testing.
                deferred = treq.put(url=url, data=logfile,
                                    headers={"Content-Type": ["text/csv"]},
                                    persistent=False)
                deferred.addCallback(lambda x: result_callback(
                    url, log_identifier, x))
                deferred.addErrback(lambda x: error_callback(
                    url, log_identifier, x))

        def result_callback(url, log_identifier, response):
            if 500 <= response.code < 600:
                delay = http_retry_delay()
                logger.error(
                    "Server side error while uploading log file %s, "
                    "status code: %s. Retrying. in %s seconds",
                    log_identifier, response.code, delay)
                upload(url, log_identifier, delay=delay)

            elif response.code not in [OK, CREATED, ACCEPTED]:
                # Nothing else we could do about that, this is
                # a problem on our end.
                logger.error(
                    "Could not upload logfile %s status code: %s. "
                    "This is a client side error, giving up.",
                    log_identifier, response.code)
                try:
                    upload_deferred.errback(ValueError(
                        "Bad return code on uploading logfile: %s"
                        % response.code))
                except Exception as e:
                    logger.error(
                        "Caught exception calling upload_deferred.errback: %s",
                        e)

            else:
                logger.info("Uploaded logfile %s for to master",
                            log_identifier)
                try:
                    upload_deferred.callback(None)
                except Exception as e:
                    logger.error(
                        "Caught exception calling upload_deferred.callback: %s",
                        e)

        def error_callback(url, log_identifier, failure_reason):
            delay = http_retry_delay()
            logger.error(
                "Error while uploading logfile %s to master: "
                "%r, retrying in %s seconds.",
                log_identifier, failure_reason, delay)
            upload(url, log_identifier, delay=delay)

        logger.info("Uploading log file %s to master, URL %r", log_identifier,
                    url)
        upload(url, log_identifier)
        return upload_deferred


class System(object):
    # complete coverage provided by other tests
    def _get_uid_gid_value(self, value, value_name, func_name,
                           module, module_name):  # pragma: no cover
        """
        Internal function which handles both user name and group conversion.
        """
        if not isinstance(value, STRING_TYPES):
            raise TypeError("Expected string for `value`")

        # This platform does not implement the module
        if module is NotImplemented:
            logger.warning(
                "This platform does not implement the %r module, skipping "
                "%s()", module_name, func_name)

        # Convert a user/group string to an integer
        elif isinstance(value, STRING_TYPES):
            try:
                if module_name == "pwd":
                    return pwd.getpwnam(value).pw_uid
                elif module_name == "grp":
                    return grp.getgrnam(value).gr_gid
                else:
                    raise ValueError(
                        "Internal error, failed to get module to use for "
                        "conversion.  Was given %r" % module)
            except KeyError:
                logger.error(
                    "Failed to convert %s to a %s",
                    value, func_name.split("_")[1])

                if not config.get("jobtype_ignore_id_mapping_errors"):
                    raise

        # Verify that the provided user/group string is real
        elif isinstance(value, INTEGER_TYPES):
            try:
                if module_name == "pwd":
                    pass
                elif module_name == "grp":
                    pass
                else:
                    raise ValueError(
                        "Internal error, failed to get module to use for "
                        "conversion.  Was given %r" % module)

                # Seems to check out, return the original value
                return value
            except KeyError:
                logger.error(
                    "%s %s does not seem to exist", value_name, value)

                if not config.get("jobtype_ignore_id_mapping_errors"):
                    raise
        else:
            raise ValueError(
                "Expected an integer or string for `%s`" % value_name)

    def _remove_tempdirs(self, max_attempts=5):
        """
        Iterates over all temporary directories in ``_tempdirs`` and removes
        them from disk.  This work will be done in a thread so it does not
        block the reactor.
        """
        def rmdir(path, on_exit_handler=True):
            logger.debug("Removing directory %s", path)

            attempts = max_attempts
            while attempts:
                try:
                    shutil.rmtree(path)
                except OSError as e:
                    attempts -= 1
                    if e.errno == ENOENT:
                        break

                    elif e.errno != ENOENT:
                        logger.error(
                            "Failed to delete %s: %s (%s attempts remain)",
                            path, e, attempts)
                else:
                    logger.debug("Removed directory %s", path)
                    break

            else:
                # If break was never called then retry on shutdown.
                if on_exit_handler:
                    atexit.register(rmdir, path, on_exit_handler=False)
                    logger.error(
                        "Failed to delete %s, will retry on shutdown", path)
                else:
                    logger.error(
                        "Failed to delete %s, no further attempts will be "
                        "made", path)

        # Delete each directory in a thread
        for directory in self._tempdirs:
            reactor.callInThread(rmdir, directory)


class TypeChecks(object):
    def _check_expandvars_inputs(self, value, environment):
        """Checks input arguments for :meth:`expandvars`"""
        if not isinstance(value, STRING_TYPES):
            raise TypeError("Expected a string for `value`")

        if environment is not None and not isinstance(environment, dict):
            raise TypeError("Expected None or a dictionary for `environment`")

    def _check_map_path_inputs(self, path):
        """Checks input arguments for :meth:`map_path`"""
        if not isinstance(path, STRING_TYPES):
            raise TypeError("Expected string for `path`")

    def _check_csvlog_path_inputs(self, protocol_uuid, now):
        """Checks input arguments for :meth:`get_csvlog_path`"""
        if not isinstance(protocol_uuid, UUID):
            raise TypeError("Expected UUID for `protocol_uuid`")

        if now is not None and not isinstance(now, datetime):
            raise TypeError("Expected None or datetime for `now`")

    def _check_command_list_inputs(self, cmdlist):
        """Checks input arguments for :meth:`get_command_list`"""
        if not isinstance(cmdlist, (tuple, list)):
            raise TypeError("Expected tuple or list for `cmdlist`")

    def _check_set_states_inputs(self, tasks, state):
        """Checks input arguments for :meth:`set_states`"""
        if not isinstance(tasks, ITERABLE_CONTAINERS):
            raise TypeError("Expected tuple, list or set for `tasks`")

        if state not in WorkState:
            raise ValueError("Expected `state` to be in %s" % list(WorkState))



# No shebang line, this module is meant to be imported
#
# Copyright 2014 Ambient Entertainment GmbH & Co. KG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import signal
import subprocess
import sys
import time
import shutil
import zipfile
from functools import partial
from os.path import join, isdir, isfile

# Platform specific imports.  These should either all fail or
# import without problems so we're grouping them together.
try:
    from grp import getgrgid
    from pwd import getpwuid
    from os import setuid, getuid, setgid, getgid, fork
except ImportError:  # pragma: no cover
    getgrgid = NotImplemented
    getpwuid = NotImplemented
    setuid = NotImplemented
    getuid = NotImplemented
    setgid = NotImplemented
    getgid = NotImplemented

from pyfarm.core.enums import INTEGER_TYPES, OS
from pyfarm.agent.config import config
from pyfarm.agent.entrypoints.parser import AgentArgumentParser
from pyfarm.agent.entrypoints.utility import start_daemon_posix
from pyfarm.agent.logger import getLogger

logger = getLogger("agent.supervisor")


def supervisor():
    logger.debug("Supervisor called with: %r", sys.argv)
    supervisor_args = []
    agent_args = []
    in_agent_args = False
    tmp_argv = list(sys.argv)
    del tmp_argv[0]
    for arg in tmp_argv:
        if not in_agent_args and arg != "--":
            supervisor_args.append(arg)
        elif not in_agent_args and arg == "--":
            in_agent_args = True
        else:
            agent_args.append(arg)

    logger.debug("supervisor_args: %s", supervisor_args)

    parser = AgentArgumentParser(
        description="Start and monitor the agent process")
    parser.add_argument("--updates-drop-dir",
                        config="agent_updates_dir",
                        type=isdir, type_kwargs=dict(create=True),
                        help="Where to look for agent updates")
    parser.add_argument("--agent-package-dir",
                        type=isdir, type_kwargs=dict(create=True),
                        help="Path to the actual agent code")
    parser.add_argument("--pidfile", config="supervisor_lock_file",
                        help="The file to store the process id in. "
                             "[default: %(default)s]")
    parser.add_argument("-n", "--no-daemon", default=False, action="store_true",
                        config=False,
                        help="If provided then do not run the process in the "
                             "background.")
    parser.add_argument("--chdir", config="agent_chdir", type=isdir,
                        help="The directory to chdir to upon launch.")
    parser.add_argument("--uid", type=int,
                        help="The user id to run the supervisor as.  "
                             "*This setting is ignored on Windows.*")
    parser.add_argument("--gid", type=int,
                        help="The group id to run the supervisor as.  "
                             "*This setting is ignored on Windows.*")
    args = parser.parse_args(supervisor_args)

    if not args.no_daemon and fork is not NotImplemented:
        logger.info("sending supervisor log output to %s" % args.log)
        daemon_start_return_code = start_daemon_posix(
            args.log, args.chdir, args.uid, args.gid)

        if isinstance(daemon_start_return_code, INTEGER_TYPES):
            return daemon_start_return_code

    elif not args.no_daemon and fork is NotImplemented:
        logger.warning(
            "`fork` is not implemented on %s, starting in "
            "foreground" % OS.title())
    else:
        logger.debug("Not forking to background")

    pid = os.getpid()
    # Write the PID file
    try:
        with open(args.pidfile, "w") as pidfile:
            pidfile.write(str(os.getpid()))
    except OSError as e:
        logger.error(
            "Failed to write PID file %s: %s", args.pidfile, e)
        return 1
    else:
        logger.debug("Wrote PID to %s", args.pidfile)

    logger.info("supervisor pid: %s" % pid)

    if getuid is not NotImplemented:
        logger.info("uid: %s" % getuid())

    if getgid is not NotImplemented:
        logger.info("gid: %s" % getgid())

    def terminate_handler(*_):
        subprocess.call(["pyfarm-agent"] + agent_args + ["stop"])
        sys.exit(0)

    def restart_handler(*_):
        subprocess.call(["pyfarm-agent"] + agent_args + ["stop"])

    logger.debug("Setting signal handlers")

    signal.signal(signal.SIGTERM, terminate_handler)
    signal.signal(signal.SIGINT, terminate_handler)
    signal.signal(signal.SIGHUP, restart_handler)

    update_file_path = join(config["agent_updates_dir"], "pyfarm-agent.zip")

    loop_interval = config["supervisor_interval"]

    while True:
        if subprocess.call(["pyfarm-agent", "status"]) != 0:
            logger.info("pyfarm-agent is not running")
            if (os.path.isfile(update_file_path) and
                zipfile.is_zipfile(update_file_path)):
                logger.info("Found an upgrade to pyfarm-agent")
                try:
                    shutil.rmtree(args.agent_package_dir)
                    os.makedirs(args.agent_package_dir)
                    with zipfile.ZipFile(update_file_path, "r") as archive:
                        archive.extractall(args.agent_package_dir)
                    os.remove(update_file_path)
                except Exception as e:
                    logger.error(
                        "Caught exception trying to update agent: %r", e)

            logger.info("starting pyfarm-agent now")
            if subprocess.call(["pyfarm-agent"] + agent_args + ["start"]) != 0:
                logger.error("Could not start pyfarm-agent")
                sys.exit(1)

        time.sleep(loop_interval)



# No shebang line, this module is meant to be imported
#
# Copyright 2013 Oliver Palmer
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Tasks
-----

Simple tasks which are run at a scheduled interval by
:class:`.ScheduledTaskManager`
"""


from twisted.internet.task import LoopingCall

from pyfarm.agent.logger import getLogger

logger = getLogger("agent.tasks")


class ScheduledTaskManager(object):
    """
    Manages and keeps track of several scheduled tasks.
    """
    test_clock = None

    def __init__(self):
        self.tasks = {}

    def register(self, function, interval, start=False, clock=None,
                 func_args=None, func_kwargs=None):
        """
        Register a callable function to run at a given interval.  This function
        will do nothing if ``function`` has already been registered.

        :param function:
            a callable function that should be run on an interval

        :type interval: int or float
        :param interval:
            the interval in which ``function`` should be urn

        :param bool start:
            if True, start the interval timer after it has been added

        :param clock:
            optional keyword that will replace the looping call's clock

        :param tuple func_args:
            the positional arguments to pass into ``function``

        :param dict func_kwargs:
            the keyword arguments to pass into ``function``

        :exception AssertionError:
            raised if ``function`` is not callable
        """
        args = func_args or ()
        kwargs = func_kwargs or {}
        assert callable(function)
        assert isinstance(args, (tuple, list))
        assert isinstance(kwargs, dict)

        if function not in self.tasks:
            looping_call = LoopingCall(function, *args, **kwargs)

            if clock is not None:
                looping_call.clock = clock

            self.tasks[function] = (looping_call, interval)
            logger.info("Registering task %r", function)

            if start:
                looping_call.start(interval)

            return looping_call

    def start(self, now=True):
        """
        start all :class:`.LoopingCall` instances stored from
        :meth:`.register`
        """
        logger.info("Starting scheduled tasks")

        for function, (looping_call, interval) in self.tasks.iteritems():
            if not looping_call.running:
                looping_call.start(interval, now=now)

    def stop(self):
        """
        stop all :class:`.LoopingCall` instances stored from
        :meth:`.register`
        """
        logger.info("Stopping scheduled tasks")

        for function, (looping_call, interval) in self.tasks.iteritems():
            if looping_call.running:
                looping_call.stop()



# No shebang line, this module is meant to be imported
#
# Copyright 2014 Oliver Palmer
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import deque
from datetime import datetime
from errno import EEXIST
from threading import Lock, RLock
from os import makedirs
from os.path import dirname, isfile

try:
    range_ = xrange
except NameError:  # pragma: no cover
    range_ = range

from twisted.internet import reactor
from twisted.internet.threads import deferToThreadPool
from twisted.python.threadpool import ThreadPool

from pyfarm.agent.config import config
from pyfarm.agent.sysinfo import cpu
from pyfarm.agent.logger import getLogger
from pyfarm.agent.utility import UnicodeCSVWriter

STDOUT = 0
STDERR = 1
STREAMS = set([STDOUT, STDERR])
CREATE_LOG_LOCK = Lock()

logger = getLogger("jobtypes.log")


def open_log(path, ignore_existing=False):
    """
    Creates a log file on disk as safely as possible and returns
    the file object.  This classmethod is meant to be run from
    within a thread so the disk IO can be placed outside of the
    reactor's event loop.

    :raise OSError:
        Raised if ``path`` already exists or its parent
        directory could not be created for some reason.
    """
    parent_dir = dirname(path)

    with CREATE_LOG_LOCK:
        if not ignore_existing and isfile(path):
            raise OSError("Log exists: %r" % path)

        # Create the directory and raise any exception
        # produced (except EEXIST)
        try:
            makedirs(parent_dir)
            logger.debug("Created directory %r", parent_dir)
        except OSError as e:  # pragma: no cover
            if e.errno != EEXIST:
                raise

        logger.debug("Opening log file %r", path)
        return open(path, "w")


class CSVLog(object):
    """
    Internal class which represents a log object and handle
    data that's internal to log handling.
    """
    def __init__(self, fileobj):
        # Check the type explicit here because UnicodeCSVWriter will
        # happy accept a non-file object the fail later.
        if not isinstance(fileobj, file):
            raise TypeError("`fileobj` should be a file")

        self.lock = RLock()
        self.file = fileobj
        self.messages = deque()
        self.lines = 0
        self.written = 0
        self.csv = UnicodeCSVWriter(self.file)

    def write(self, data):
        """Writes the given data to the underlying csv object."""
        date, streamno, lineno, message = data
        self.csv.writerow(
            [date.isoformat(), str(streamno), str(lineno), message])
        self.written += 1


class LoggerPool(ThreadPool):
    """
    This is the thread pool which will handle create of log file
    and storage of messages from processes.  Configuration of
    some of the internals of this class can accomplished using the
    ``jobtype_logging_threadpool`` config value in ``jobtypes.yml``.
    """
    reactor = reactor
    logs = {}

    def __init__(self):
        minthreads = config["jobtype_logging_threadpool"]["min_threads"]
        maxthreads = config["jobtype_logging_threadpool"]["max_threads"]
        self.max_queued_lines = \
            config["jobtype_logging_threadpool"]["max_queue_size"]
        self.flush_lines = \
            config["jobtype_logging_threadpool"]["flush_lines"]
        self.stopped = False

        if minthreads < 1:
            raise ValueError(
                "Config value "
                "jobtype_logging_threadpool.min_threads must be >= 1")

        # Calculate maxthreads if a value was not provided for us
        if maxthreads == "auto":
            auto_maxthreads = min(int(cpu.total_cpus() * 1.5), 20)
            maxthreads = max(auto_maxthreads, minthreads)

        if minthreads > maxthreads:
            raise ValueError(
                "Config value jobtype_logging_threadpool.min_threads cannot "
                "be larger than jobtype_logging_threadpool.max_threads")

        ThreadPool.__init__(
            self,
            minthreads=minthreads, maxthreads=maxthreads,
            name=self.__class__.__name__)

    def defer(self, function, *args, **kwargs):  # pragma: no cover
        """
        Wrapper around :func:`.deferToThreadPool` so we can defer some
        functions to this pool's workers.
        """
        return deferToThreadPool(self.reactor, self, function, *args, **kwargs)

    def open_log(self, protocol, path, ignore_existing=False):
        """
        Opens a log file for the given ``protocol`` object at ``path``.  This
        class method will return a deferred object that will fire its callback
        once the log is open and ready to receive data.
        """
        if protocol.uuid in self.logs:
            raise KeyError(
                "Protocol %r is already logging to %r" % (
                    protocol.uuid, self.logs[protocol.uuid]))

        def log_created(stream, impacted_protocol):
            logger.info(
                "Created log for protocol %r at %r",
                impacted_protocol.uuid, stream.name)

            self.logs[impacted_protocol.uuid] = CSVLog(stream)
            return impacted_protocol.uuid, self.logs[impacted_protocol.uuid]

        deferred = self.defer(open_log, path, ignore_existing=ignore_existing)
        deferred.addCallback(log_created, protocol)
        return deferred

    def close_log(self, protocol_uuid):
        """Closes the file handle for the given protocol id."""
        log = self.logs.pop(protocol_uuid, None)
        if log is not None:
            self.flush(log)
            log.file.close()
            logger.info("Closed %s", log.file.name)

    def log(self, protocol_id, streamno, message):
        """
        Places a single message to be handled by the worker threads into
        the queue for processing.
        """
        if self.stopped:
            logger.warning(
                "Rejecting log message, pool is stopped or stopping!!")
            return

        # This operation is atomic so we're safe to keep
        log = self.logs[protocol_id]
        log.lines += 1
        log.messages.append((datetime.utcnow(), streamno, log.lines, message))

        if len(log.messages) > self.max_queued_lines:
            self.callInThread(self.flush, log)

    def flush(self, log):
        """
        Takes the given log object and flushes the messages it
        contains to the attached file object.
        """
        while True:
            # Only one thread at a time may retrieve objects, write
            # to the file, and flush.  This helps to preserve the
            # order of the messages and cuts down on wasted cycles
            # from switching contexts.
            with log.lock:
                try:
                    data = log.messages.popleft()
                except IndexError:
                    break
                else:
                    try:
                        log.write(data)
                    except (OSError, IOError) as e:  # pragma: no cover
                        # Put the log message back in the queue
                        # so we're not losing data.  It may be lightly
                        # out of order now but we have a date stamp
                        # and it's more important we don't lose data.
                        log.messages.appendleft(data)
                        logger.error(
                            "Failed to write to %s: %s", log.file.name, e)

        # Check if we should flush to disk.  We're doing
        # this outside the above because it ensures we
        # only have to run this logic once instead of
        # once per message.
        with log.lock:
            if log.written >= self.flush_lines:
                try:
                    log.file.flush()
                except (OSError, IOError) as e:  # pragma: no cover
                    logger.error(
                        "Failed to flush output to %s: %s",
                        log.file.name, e)
                else:
                    logger.debug(
                        "%s wrote %s lines to %s",
                        self.currentThread().name, log.written,
                        log.file.name)
                    log.written = 0

        return log

    def stop(self):
        """
        Flushes any remaining data, closes the underlying files, then stops
        the thread pool.

        .. warning::

            Because this method is usually called when the reactor is
            stopping all file handling happens in the main thread.
        """
        logger.info("Logging thread pool is shutting down.")
        self.stopped = True

        for protocol_id in list(self.logs.keys()):
            self.close_log(protocol_id)

        ThreadPool.stop(self)

    def start(self):
        """
        Starts the logger pool and sets up the required shutdown event
        trigger which will call :meth:`stop` on exit.
        """
        reactor.addSystemEventTrigger("before", "shutdown", self.stop)
        ThreadPool.start(self)
        logger.info(
            "Started logger thread pool (min=%s, max=%s)", self.min, self.max)

try:
    logpool
    logger.warning(  # pragma: no cover
        "A logger pool was already setup, seems like the log module has "
        "been imported more than once.")
except NameError:
    logpool = LoggerPool()

    # Required so we only start the logpool if the
    # reactor is going to run.  If we don't do this
    # other things will hang.
    reactor.callWhenRunning(logpool.start)



# No shebang line, this module is meant to be imported
#
# Copyright 2014 Oliver Palmer
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Python Logger
-------------

This module provides the facilities to capture and send
log records from Python's logger into Twisted.  It also
provides a :class:`Logger` class and :func:`getLogger`
function to replace the built-in Python implementations.
"""

from time import time
from logging import (
    NOTSET, DEBUG, INFO, WARNING, ERROR, CRITICAL, FATAL, Handler)

from twisted.python.log import msg


class Logger(object):
    """
    A stand-in for an instance of :class:`logging.Logger`
    Unlike the standard logger this just forwards all messages
    to Twisted's logging system.
    """
    def __init__(self, name):
        self.name = name

    def debug(self, message, *args):
        msg(message, args=args, system=self.name,
            time=time(), logLevel=DEBUG)

    def info(self, message, *args):
        msg(message, args=args, system=self.name,
            time=time(), logLevel=INFO)

    def warning(self, message, *args):
        msg(message, args=args, system=self.name,
            time=time(), logLevel=WARNING)

    def error(self, message, *args):
        msg(message, args=args, system=self.name,
            time=time(), logLevel=ERROR)

    def critical(self, message, *args):
        msg(message, args=args, system=self.name,
            time=time(), logLevel=CRITICAL)

    def fatal(self, message, *args):
        msg(message, args=args, system=self.name,
            time=time(), logLevel=FATAL)


class LogRecordToTwisted(Handler):
    """
    Captures logging events for a standard Python logger
    and sends them to Twisted.  Twisted has a built in
    class to help work with Python's logging library
    however it won't translate everything directly.
    """
    def __init__(self):
        # We don't use these attributes because the observer
        # handles these.  But we still have to provide them
        # because the Python logger system will try to access
        # them.
        self.level = NOTSET
        self.filters = []

    def emit(self, record):
        """
        Emits an instance of :class:`logging.LogRecord` into Twisted's
        logging system.
        """
        msg(record.msg, args=record.args, python_record=True,
            time=record.created, system=record.name, logLevel=record.levelno)

    def acquire(self):
        pass

    def release(self):
        pass

    def createLock(self):
        pass

    def close(self):
        pass


def getLogger(name):
    """
    Analog to Python's :func:`logging.getLogger` except it
    returns instances of :class:`Logger` instead.
    """
    return Logger("pf.%s" % name)


# No shebang line, this module is meant to be imported
#
# Copyright 2014 Oliver Palmer
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Logger
======

This module provides the functionality for handling log
messages in the agent.
"""

from logging import DEBUG, getLogger as _getLogger

from twisted.python.log import startLoggingWithObserver

from pyfarm.agent.logger.twistd import Observer
from pyfarm.agent.logger.python import LogRecordToTwisted, getLogger

# NOTE: Normally we don't place code in __init__.  In this case we're doing
# so because we don't want to change the existing interface for getLogger
# and we only want to expose two functions.

__all__ = ["setup_logging"]


def setup_logging():
    """
    Sets up Twisted's and Python's logging system
    to run through our observer.  This ensures that all
    log messages are handled in a single location and
    that we don't lose any messages.
    """
    if Observer.INSTANCE is None:
        # Setup and configure the logger for Twisted
        observer = Observer()
        startLoggingWithObserver(observer, setStdout=False)

        # Ensure all of Python's log flow into Twisted
        root = _getLogger("")
        root.setLevel(DEBUG)
        root.handlers[:] = [LogRecordToTwisted()]

        # Now that we've configured the observer import the
        # the config and configure the observer itself
        observer.configure()
        Observer.INSTANCE = observer



# No shebang line, this module is meant to be imported
#
# Copyright 2013 Oliver Palmer
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
HTTP Client
===========

The client library the manager uses to communicate with
the master server.
"""

import json
from collections import namedtuple
from functools import partial
from random import random

try:
    from httplib import responses
except ImportError:  # pragma: no cover
    from http.client import responses

try:
    from UserDict import UserDict
except ImportError:  # pragma: no cover
    from collections import UserDict

import treq
try:
    from treq.response import _Response as TQResponse
except ImportError:  # pragma: no cover
    TQResponse = NotImplemented

from twisted.internet.defer import Deferred
from twisted.internet.protocol import Protocol, connectionDone
from twisted.python import log
from twisted.web.client import (
    Response as TWResponse, GzipDecoder as TWGzipDecoder, ResponseDone)

from pyfarm.core.enums import STRING_TYPES, NOTSET, INTEGER_TYPES
from pyfarm.core.utility import ImmutableDict
from pyfarm.agent.config import config
from pyfarm.agent.logger import getLogger
from pyfarm.agent.utility import quote_url, json_safe

logger = getLogger("agent.http.client")

# response classes which are allowed to the `response` argument
# to Response.__init__
if TQResponse is not NotImplemented:
    RESPONSE_CLASSES = (TWResponse, TWGzipDecoder, TQResponse)
else:  # pragma: no cover
    RESPONSE_CLASSES = (TWResponse, TWGzipDecoder)

USERAGENT = "PyFarm/1.0 (agent)"
DELAY_NUMBER_TYPES = tuple(list(INTEGER_TYPES) + [float])


def build_url(url, params=None):
    """
    Builds the full url when provided the base ``url`` and some
    url parameters:

    >>> build_url("/foobar", {"first": "foo", "second": "bar"})
    '/foobar?first=foo&second=bar'
    >>> build_url("/foobar bar/")
    ''/foobar%20bar/'

    :param str url:
        The url to build off of.

    :param dict params:
        A dictionary of parameters that should be added on to ``url``.  If
        this value is not provided ``url`` will be returned by itself.
        Arguments to a url are unordered by default however they will be
        sorted alphabetically so the results are repeatable from call to call.
    """
    assert isinstance(url, STRING_TYPES)

    # append url arguments
    if isinstance(params, (dict, ImmutableDict, UserDict)) and params:
        url += "?" + "&".join([
            "%s=%s" % (key, value)for key, value in sorted(params.items())])

    return quote_url(url)


def http_retry_delay(initial=None, uniform=False, get_delay=random, minimum=1):
    """
    Returns a floating point value that can be used to delay
    an http request.  The main purpose of this is to ensure that not all
    requests are run with the same interval between then.  This helps to
    ensure that if the same request, such as agents coming online, is being
    run on multiple systems they should be staggered a little more than
    they would be without the non-uniform delay.

    :param int initial:
        The initial delay value to start off with before any extra
        calculations are done.  If this value is not provided the value
        provided to ``--http-retry-delay`` at startup will be used.

    :param bool uniform:
        If True then use the value produced by ``get_delay`` as
        a multiplier.

    :param callable get_delay:
        A function which should produce a number to multiply ``delay``
        by.  By default this uses :func:`random.random`

    :param minimum:
        Ensures that the value returned from this function is greater than
        or equal to a minimum value.
    """
    delay = initial
    if initial is None:
        # TODO: provide command line flags for jitter
        delay = config["agent_http_retry_delay"]

    assert isinstance(delay, DELAY_NUMBER_TYPES)  # expect integer or float
    assert isinstance(minimum, DELAY_NUMBER_TYPES)  # expect integer or float
    assert minimum >= 0

    if not uniform:
        delay *= get_delay()

    return max(minimum, delay)


class Request(namedtuple("Request", ("method", "url", "kwargs"))):
    """
    Contains all the information used to perform a request such as the
    ``method``, ``url``, and original keyword arguments (``kwargs``).  These
    values contain the basic information necessary in order to :meth:`retry`
    a request.
    """
    def retry(self, **kwargs):
        """
        When called this will rerun the original request with all
        of the original arguments to :func:`request`

        :param kwargs:
            Additional keyword arguments which should override the original
            keyword argument(s).
        """
        # first take the original keyword arguments
        # and provide overrides
        request_kwargs = self.kwargs.copy()
        request_kwargs.update(kwargs)

        # log and retry the request
        debug_kwargs = request_kwargs.copy()
        url = build_url(self.url, debug_kwargs.pop("params", None))
        logger.debug(
            "Retrying %s %s, kwargs: %r", self.method, url, debug_kwargs)
        return request(self.method, self.url, **request_kwargs)


class Response(Protocol):
    """
    This class receives the incoming response body from a request
    constructs some convenience methods and attributes around the data.

    :param Deferred deferred:
        The deferred object which contains the target callback
        and errback.

    :param response:
        The initial response object which will be passed along
        to the target deferred.

    :param Request request:
        Named tuple object containing the method name, url, headers, and data.
    """
    def __init__(self, deferred, response, request):
        assert isinstance(deferred, Deferred)
        assert isinstance(response, RESPONSE_CLASSES)
        assert isinstance(request, Request)

        # internal attributes
        self._done = False
        self._body = ""
        self._deferred = deferred

        # main public attributes
        self.request = request
        self.response = response

        # convenience attributes constructed
        # from the public attributes
        self.method = self.request.method
        self.url = self.request.url
        self.code = self.response.code
        self.content_type = None

        # consume the response headers
        self.headers = {}
        for header_key, header_value in response.headers.getAllRawHeaders():
            if len(header_value) == 1:
                header_value = header_value[0]
            self.headers[header_key] = header_value

        # determine the content type
        if "Content-Type" in self.headers:
            self.content_type = self.headers["Content-Type"]

    def data(self):
        """
        Returns the data currently contained in the buffer.

        :raises RuntimeError:
            Raised if this method id called before all data
            has been received.
        """
        if not self._done:
            raise RuntimeError("Response not yet received.")
        return self._body

    def json(self, loader=json.loads):
        """
        Returns the json data from the incoming request

        :raises RuntimeError:
            Raised if this method id called before all data
            has been received.

        :raises ValueError:
            Raised if the content type for this request is not
            application/json.
        """
        if not self._done:
            raise RuntimeError("Response not yet received.")
        elif self.content_type != "application/json":
            raise ValueError("Not an application/json response.")
        else:
            return loader(self._body)

    def dataReceived(self, data):
        """
        Overrides :meth:`.Protocol.dataReceived` and appends data
        to ``_body``.
        """
        self._body += data

    def connectionLost(self, reason=connectionDone):
        """
        Overrides :meth:`.Protocol.connectionLost` and sets the ``_done``
        when complete.  When called with :class:`.ResponseDone` for ``reason``
        this method will call the callback on ``_deferred``
        """
        if reason.type is ResponseDone:
            self._done = True
            url = build_url(self.request.url, self.request.kwargs.get("params"))
            code_text = responses.get(self.code, "UNKNOWN")
            logger.debug(
                "%s %s %s %s, body: %s",
                self.code, code_text, self.request.method, url, self._body)
            self._deferred.callback(self)
        else:
            self._deferred.errback(reason)


def request(method, url, **kwargs):
    """
    Wrapper around :func:`treq.request` with some added arguments
    and validation.

    :param str method:
        The HTTP method to use when making the request.

    :param str url:
        The url this request will be made to.

    :type data: str, list, tuple, set, dict
    :keyword data:
        The data to send along with some types of requests
        such as ``POST`` or ``PUT``

    :keyword dict headers:
        The headers to send along with the request to
        ``url``.  Currently only single values per header
        are supported.

    :keyword function callback:
        The function to deliver an instance of :class:`Response`
        once we receive and unpack a response.

    :keyword function errback:
        The function to deliver an error message to.  By default
        this will use :func:`.log.err`.

    :keyword class response_class:
        The class to use to unpack the internal response.  This is mainly
        used by the unittests but could be used elsewhere to add some
        custom behavior to the unpack process for the incoming response.
    """
    # check assumptions for arguments
    assert method in ("HEAD", "GET", "POST", "PUT", "PATCH", "DELETE")
    assert isinstance(url, STRING_TYPES) and url

    original_request = Request(
        method=method, url=url, kwargs=ImmutableDict(kwargs.copy()))
    data = kwargs.pop("data", NOTSET)
    headers = kwargs.pop("headers", {})
    callback = kwargs.pop("callback", None)
    errback = kwargs.pop("errback", log.err)
    response_class = kwargs.pop("response_class", Response)

    # check assumptions for keywords
    assert callback is not None, "callback not provided"
    assert callable(callback) and callable(errback)
    assert data is NOTSET or \
           isinstance(data, tuple(list(STRING_TYPES) + [dict, list]))

    # add our default headers
    headers.setdefault("Content-Type", ["application/json"])
    headers.setdefault("User-Agent", [USERAGENT])

    # ensure all values in the headers are lists (needed by Twisted)
    for header, value in headers.items():
        if isinstance(value, STRING_TYPES):
            headers[header] = [value]

        # for our purposes we should not expect headers with more
        # than one value for now
        elif isinstance(value, (list, tuple, set)):
            assert len(value) == 1

        else:
            raise NotImplementedError(
                "cannot handle header values with type %s" % type(value))

    def unpack_response(response):
        deferred = Deferred()
        deferred.addCallback(callback)

        # Deliver the body onto an instance of the response
        # object along with the original request.  Finally
        # the request and response via an instance of `Response`
        # to the outer scope's callback function.
        response.deliverBody(
            response_class(deferred, response, original_request))

        return deferred

    # prepare to send the data
    if isinstance(data, dict):
        data = json_safe(data)

    if data is not NOTSET and \
                    headers["Content-Type"] == ["application/json"]:
        data = json.dumps(data)

    elif data is not NOTSET:
        raise NotImplementedError(
            "Don't know how to dump data for %s" % headers["Content-Type"])

    # prepare keyword arguments
    kwargs.update(
        headers=headers,
        persistent=config["agent_http_persistent_connections"])

    if data is not NOTSET:
        kwargs.update(data=data)

    debug_kwargs = kwargs.copy()
    debug_url = build_url(url, debug_kwargs.pop("params", None))
    logger.debug(
        "Queued %s %s, kwargs: %r", method, debug_url, debug_kwargs)

    try:
        deferred = treq.request(method, quote_url(url), **kwargs)
    except NotImplementedError:  # pragma: no cover
        logger.error(
            "Attempting to access a url over SSL but you don't have the "
            "proper libraries installed.  Please install the PyOpenSSL and "
            "service_identity Python packages.")
        raise

    deferred.addCallback(unpack_response)
    deferred.addErrback(errback)

    return deferred


head = partial(request, "HEAD")
get = partial(request, "GET")
post = partial(request, "POST")
put = partial(request, "PUT")
patch = partial(request, "PATCH")
delete = partial(request, "DELETE")



# No shebang line, this module is meant to be imported
#
# Copyright 2014 Ambient Entertainment GmbH & Co. KG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from json import dumps

try:
    from httplib import BAD_REQUEST, OK
except ImportError:  # pragma: no cover
    from http.client import BAD_REQUEST, OK

from os.path import join

from twisted.web.server import NOT_DONE_YET
from twisted.protocols.basic import FileSender

from pyfarm.agent.config import config
from pyfarm.agent.http.api.base import APIResource
from pyfarm.agent.utility import request_from_master


class TaskLogs(APIResource):
    def get(self, **kwargs):
        """
        Get the contents of the specified task log
        """
        request = kwargs["request"]

        if request_from_master(request):
            config.master_contacted()

        if len(request.postpath) != 1:
            request.setResponseCode(BAD_REQUEST)
            request.write(dumps({"error": "Did not specify a log identifier"}))

        log_identifier = request.postpath[0]
        if "/" in log_identifier or "\\" in log_identifier:
            request.setResponseCode(BAD_REQUEST)
            request.write(dumps({"error": "log_identifier must not contain"
                                          "directory separators"}))

        path = join(config["jobtype_task_logs"], log_identifier)
        request.setResponseCode(OK)
        request.setHeader("Content-Type", "text/csv")

        logfile = open(path, "rb")
        deferred = FileSender().beginFileTransfer(logfile, request)
        def transfer_finished(_):
            request.finish()
            logfile.close()
        deferred.addCallback(transfer_finished)

        return NOT_DONE_YET



# No shebang line, this module is meant to be imported
#
# Copyright 2014 Oliver Palmer
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Template
--------

Interface methods for working with the Jinja template
engine.
"""

from io import BytesIO

from jinja2 import (
    Environment as _Environment, Template, PackageLoader, BytecodeCache)
from twisted.internet.defer import Deferred

from pyfarm.agent.config import config


class InMemoryCache(BytecodeCache):
    """
    Caches Jinja templates into memory after they have been loaded
    and compiled.
    """
    cache = {}

    def clear(self):
        self.cache.clear()

    # Untested as this is an internal implementation for
    # the Environment class below.
    def load_bytecode(self, bucket):  # pragma: no cover
        if bucket.key in self.cache:
            bucket.load_bytecode(self.cache[bucket.key])

    # Untested as this is an internal implementation for
    # the Environment class below.
    def dump_bytecode(self, bucket):  # pragma: no cover
        cache = BytesIO()
        bucket.write_bytecode(cache)
        self.cache[bucket.key] = cache


class DeferredTemplate(Template):
    """
    Overrides the default :class:`.PackageLoader` so we
    can produced the rendered result as a deferred call.
    """
    def render(self, *args, **kwargs):
        deferred = Deferred()

        try:
            # get the results then convert to a string, Twisted can't handle
            # unicode from here
            deferred.callback(
                str(super(DeferredTemplate, self).render(*args, **kwargs)))
        except Exception as e:  # pragma: no cover
            deferred.errback(e)

        return deferred


class Environment(_Environment):
    """
    Implementation of Jinja's :class:`._Environment` class which
    reads from our configuration object and establishes the
    default functions we can use in a template.
    """
    template_class = DeferredTemplate

    def __init__(self, **kwargs):
        # default options
        kwargs.setdefault("bytecode_cache", InMemoryCache())
        kwargs.setdefault("loader", PackageLoader("pyfarm.agent.http"))
        kwargs.setdefault("auto_reload", config["agent_html_template_reload"])

        super(Environment, self).__init__(**kwargs)

        # global functions which are available within
        # the templates
        self.globals.update(
            is_int=lambda value: isinstance(value, int),
            is_str=lambda value: isinstance(value, (str, unicode)),
            typename=lambda value: type(value).__name__,
            agent_hostname=lambda: config["agent_hostname"],
            agent_id=lambda: config["agent-id"],
            state=lambda: config["state"],
            repr=repr)


class Loader(object):
    """
    Namespace class used to simply keep track of the global environment
    and load templates.

    >>> from pyfarm.agent.http.core import template
    >>> template.load("index.html")
    """
    environment = None

    @classmethod
    def load(cls, name):
        if cls.environment is None:
            cls.environment = Environment()

        return cls.environment.get_template(name)

load = Loader.load



# No shebang line, this module is meant to be imported
#
# Copyright 2013 Oliver Palmer
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
CPU
---

Contains information about the cpu and its relation to the operating
system such as load, processing times, etc.
"""

from __future__ import division

import psutil


def total_cpus(logical=True):
    """
    Returns the total number of cpus installed on the system.

    :param bool logical:
        If True the return the number of cores the system has.  Setting
        this value to False will instead return the number of physical
        cpus present on the system.
    """
    return psutil.cpu_count(logical=logical)


def load(interval=1):
    """
    Returns the load across all cpus value from zero to one.  A value
    of 1.0 means the average load across all cpus is 100%.
    """
    return psutil.cpu_percent(interval) / total_cpus()


def user_time():
    """
    Returns the amount of time spent by the cpu in user
    space
    """
    return psutil.cpu_times().user


def system_time():
    """
    Returns the amount of time spent by the cpu in system
    space
    """
    return psutil.cpu_times().system


def idle_time():
    """
    Returns the amount of time spent by the cpu in idle
    space
    """
    return psutil.cpu_times().idle


def iowait():
    """
    Returns the amount of time spent by the cpu waiting
    on io

    .. note::
        on platforms other than linux this will return None
    """
    try:
        cpu_times = psutil.cpu_times()
        if hasattr(cpu_times, "iowait"):
            return psutil.cpu_times().iowait
    except AttributeError:  # pragma: no cover
        return None


# No shebang line, this module is meant to be imported
#
# Copyright 2014 Oliver Palmer
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Utilities
---------

Top level utilities for the agent to use internally.  Many of these
are copied over from the master (which we can't import here).
"""

import csv
import codecs
import cStringIO
from decimal import Decimal
from datetime import datetime, timedelta
from json import dumps as _dumps
from UserDict import UserDict
from uuid import UUID, uuid1

try:
    from urlparse import urlsplit
    from urllib import quote
except ImportError:  # pragma: no cover
    from http.client import urlsplit
    from urllib.parse import quote

try:
    from httplib import OK
except ImportError:  # pragma: no cover
    from http.client import OK

from voluptuous import Schema, Any, Required, Optional, Invalid

from pyfarm.core.enums import STRING_TYPES
from pyfarm.agent.config import config
from pyfarm.agent.logger import getLogger

logger = getLogger("agent.util")
STRINGS = Any(*STRING_TYPES)
try:
    WHOLE_NUMBERS = Any(*(int, long))
    NUMBERS = Any(*(int, long, float, Decimal))
except NameError:  # pragma: no cover
    WHOLE_NUMBERS = int
    NUMBERS = Any(*(int, float, Decimal))


def validate_environment(values):
    """
    Ensures that ``values`` is a dictionary and that it only
    contains string keys and values.
    """
    if not isinstance(values, dict):
        raise Invalid("Expected a dictionary")

    for key, value in values.items():
        if not isinstance(key, STRING_TYPES):
            raise Invalid("Key %r must be a string" % key)

        if not isinstance(value, STRING_TYPES):
            raise Invalid("Value %r for key %r must be a string" % (key, value))


def validate_uuid(value):
    """
    Ensures that ``value`` can be converted to or is a UUID object.
    """
    if isinstance(value, UUID):
        return value
    elif isinstance(value, STRING_TYPES):
        try:
            return UUID(hex=value)
        except ValueError:
            raise Invalid("%s cannot be converted to a UUID" % value)
    else:
        raise Invalid("Expected string or UUID instance for `value`")


# Shared schema declarations
JOBTYPE_SCHEMA = Schema({
    Required("name"): STRINGS,
    Required("version"): WHOLE_NUMBERS})
TASK_SCHEMA = Schema({
    Required("id"): WHOLE_NUMBERS,
    Required("frame"): NUMBERS,
    Required("attempt", default=0): WHOLE_NUMBERS})
TASKS_SCHEMA = lambda values: map(TASK_SCHEMA, values)
JOB_SCHEMA = Schema({
    Required("id"): WHOLE_NUMBERS,
    Required("by"): NUMBERS,
    Optional("batch"): WHOLE_NUMBERS,
    Optional("user"): STRINGS,
    Optional("ram"): WHOLE_NUMBERS,
    Optional("ram_warning"): WHOLE_NUMBERS,
    Optional("ram_max"): WHOLE_NUMBERS,
    Optional("cpus"): WHOLE_NUMBERS,
    Optional("data"): dict,
    Optional("environ"): validate_environment,
    Optional("title"): STRINGS})


def uuid():
    """Wrapper around :func:`uuid1` which incorporates our system id"""
    return uuid1(node=config["agent_systemid"])


def default_json_encoder(obj, return_obj=False):
    if isinstance(obj, Decimal):
        return float(obj)
    elif isinstance(obj, datetime):
        return obj.isoformat()
    elif isinstance(obj, UUID):
        return str(obj)
    elif return_obj:
        return obj


def json_safe(source):
    """
    Recursively converts ``source`` into something that should be safe for
    :func:`json.dumps` to handle.  This is used in conjunction with
    :func:`default_json_encoder` to also convert keys to something the json
    encoder can understand.
    """
    if not isinstance(source, dict):
        return source

    result = {}

    try:
        items = source.iteritems
    except AttributeError:  # pragma: no cover
        items = source.items

    for k, v in items():
        result[default_json_encoder(k, return_obj=True)] = \
            default_json_encoder(json_safe(v), return_obj=True)

    return result


def quote_url(source_url):
    """
    This function serves as a wrapper around :func:`.urlsplit`
    and :func:`.quote` and a url that has the path quoted.
    """
    url = urlsplit(source_url)

    # If a url is just "/" then we don't want to add ://
    # since a scheme was not included.
    if url.scheme:
        result = "{scheme}://".format(scheme=url.scheme)
    else:
        result = ""

    result += url.netloc + quote(url.path, safe="/?&=")

    if url.query:
        result += "?" + url.query

    if url.fragment:
        result += "#" + url.fragment

    return result


def dumps(*args, **kwargs):
    """
    Agent's implementation of :func:`json.dumps` or
    :func:`pyfarm.master.utility.jsonify`
    """
    indent = None
    if config["agent_pretty_json"]:
        indent = 2

    if len(args) == 1 and not isinstance(args[0], (dict, UserDict)):
        obj = args[0]
    else:
        obj = dict(*args, **kwargs)

    return _dumps(obj, default=default_json_encoder, indent=indent)


def request_from_master(request):
    """Returns True if the request appears to be coming from the master"""
    return request.getHeader("User-Agent") == config["master_user_agent"]


# Unicode CSV reader/writers from the standard library docs:
#   https://docs.python.org/2/library/csv.html

class UTF8Recoder(object):
    """
    Iterator that reads an encoded stream and reencodes the input to UTF-8
    """
    def __init__(self, f, encoding):
        self.reader = codecs.getreader(encoding)(f)

    def __iter__(self):
        return self

    def next(self):
        return self.reader.next().encode("utf-8")


class UnicodeCSVReader(object):
    """
    A CSV reader which will iterate over lines in the CSV file "f",
    which is encoded in the given encoding.
    """
    def __init__(self, f, dialect=csv.excel, encoding="utf-8", **kwds):
        f = UTF8Recoder(f, encoding)
        self.reader = csv.reader(f, dialect=dialect, **kwds)

    def next(self):
        row = self.reader.next()
        return [unicode(s, "utf-8") for s in row]

    def __iter__(self):
        return self


class UnicodeCSVWriter(object):
    """
    A CSV writer which will write rows to CSV file "f",
    which is encoded in the given encoding.
    """
    def __init__(self, f, dialect=csv.excel, encoding="utf-8", **kwds):
        # Redirect output to a queue
        self.queue = cStringIO.StringIO()
        self.writer = csv.writer(self.queue, dialect=dialect, **kwds)
        self.stream = f
        self.encoder = codecs.getincrementalencoder(encoding)()

    def writerow(self, row):
        self.writer.writerow([s.encode("utf-8") for s in row])
        # Fetch UTF-8 output from the queue ...
        data = self.queue.getvalue()
        data = data.decode("utf-8")
        # ... and reencode it into the target encoding
        data = self.encoder.encode(data)
        # write to the target stream
        self.stream.write(data)
        # empty queue
        self.queue.truncate(0)

    def writerows(self, rows):
        for row in rows:
            self.writerow(row)


def total_seconds(td):
    """
    Returns the total number of seconds in the time delta
    object.  This function is provided for backwards
    comparability with Python 2.6.
    """
    assert isinstance(td, timedelta)
    try:
        return td.total_seconds()
    except AttributeError:  # pragma: no cover
        return (
           td.microseconds + (td.seconds + td.days * 24 * 3600) * 1e6) / 1e6



# No shebang line, this module is meant to be imported
#
# Copyright 2014 Oliver Palmer
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
from datetime import timedelta

try:
    from httplib import OK
except ImportError:  # pragma: no cover
    from http.client import OK

import psutil
from twisted.web.server import NOT_DONE_YET
from pyfarm.core.enums import AgentState
from pyfarm.agent.config import config
from pyfarm.agent.http.core.resource import Resource
from pyfarm.agent.sysinfo import cpu, memory


def mb(value):
    if isinstance(value, float):
        value = int(value)
    return str(value) + "MB"


def seconds(value):
    if value is not None:
        return "%.2f seconds" % value


class Index(Resource):
    """serves request for the root, '/', target"""
    TEMPLATE = "index.html"

    def get(self, **kwargs):
        request = kwargs["request"]

        # write out the results from the template back
        # to the original request
        def cb(content):
            request.write(content)
            request.setResponseCode(OK)
            request.finish()

        # convert the state integer to a string
        for key, value in AgentState._asdict().iteritems():
            if config["state"] == value:
                state = key.title()
                break
        else:  # pragma: no cover
            raise KeyError("failed to find state")

        ram_allocated = int((memory.used_ram() / float(config["agent_ram"])) * 100)

        if ram_allocated >= 100:  # pragma: no cover
            ram_css = "danger"
        elif ram_allocated >= 75:  # pragma: no cover
            ram_css = "warning"
        elif ram_allocated >= 50:  # pragma: no cover
            ram_css = "info"
        else:
            ram_css = None

        memory_info = [
            ("RAM Used",
                "%.2f%% (%s of %s)" % (
                    ram_allocated,
                    memory.used_ram(),
                    mb((config["agent_ram"]))), ram_css),
            ("System RAM", memory.total_ram(), None),
            ("System RAM (reported)", mb(config["agent_ram"]), None),
            ("Agent RAM Usage", memory.process_memory(), None)]

        network_info = [
            ("Hostname", config["agent_hostname"]),
            ("Agent Port", config["agent_api_port"]),
            ("Master API", config["master_api"])]

        cpu_info = [
            ("CPUs", cpu.total_cpus()),
            ("CPUs (reported)", config["agent_cpus"]),
            ("System Time", seconds(cpu.system_time())),
            ("User Time", seconds(cpu.user_time())),
            ("Idle Time", seconds(cpu.idle_time())),
            ("IO Wait", seconds(cpu.iowait()) or "Not Supported")]

        agent_id = config["agent-id"] if "agent_id" in config else None

        miscellaneous = [
            ("Database ID", agent_id),
            ("Logged On User(s)",
             ", ".join(sorted(set(user.name for user in psutil.get_users())))),
            ("Host Uptime",
             str(timedelta(seconds=time.time() - psutil.boot_time()))),
            ("Agent Uptime",
             str(timedelta(seconds=time.time() - config["start"])))]

        deferred = self.template.render(
            memory_info=memory_info,
            network_info=network_info,
            cpu_info=cpu_info,
            miscellaneous=miscellaneous)
        deferred.addCallback(cb)

        return NOT_DONE_YET


# TODO: form submission
# TODO: make 'port' field editable (requires restart)
# TODO: add callbacks for any field that needs to update the db
class Configuration(Resource):
    TEMPLATE = "configuration.html"

    # fields which nobody can see
    HIDDEN_FIELDS = (
        "agent", "agent_pretty_json")

    # fields that a user can edit
    EDITABLE_FIELDS = (
        "agent_cpus", "agent_hostname", "agent_http_retry_delay",
        "master_api", "master", "agent_ram_check_interval", "agent_ram",
        "agent_ram_report_delta", "agent_time_offset", "state",
        "agent_http_retry_delay")

    def get(self, **kwargs):
        request = kwargs["request"]

        # write out the results from the template back
        # to the original request
        def cb(content):
            request.write(content)
            request.setResponseCode(OK)
            request.finish()

        editable_fields = []
        non_editable_fields = []
        for key, value in sorted(config.items()):
            if key in self.HIDDEN_FIELDS:
                pass
            elif key in self.EDITABLE_FIELDS:
                editable_fields.append((key, value))
            else:
                non_editable_fields.append((key, value))

        deferred = self.template.render(
            non_editable_fields=non_editable_fields,
            editable_fields=editable_fields)
        deferred.addCallback(cb)

        return NOT_DONE_YET



# No shebang line, this module is meant to be imported
#
# Copyright 2014 Oliver Palmer
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Main
----

The main module which constructs the entrypoint for the
``pyfarm-agent`` command line tool.
"""

from __future__ import division

import os
import sys
import pdb
import time

from json import dumps

from os.path import dirname, isfile, isdir

try:
    from httplib import ACCEPTED, OK, responses
except ImportError:  # pragma: no cover
    from http.client import ACCEPTED, OK, responses

# Platform specific imports.  These should either all fail or
# import without problems so we're grouping them together.
try:
    from grp import getgrgid
    from pwd import getpwuid
    from os import setuid, getuid, setgid, getgid, fork
except ImportError:  # pragma: no cover
    getgrgid = NotImplemented
    getpwuid = NotImplemented
    setuid = NotImplemented
    getuid = NotImplemented
    setgid = NotImplemented
    getgid = NotImplemented
    fork = NotImplemented

import psutil
import requests
import signal
from requests import ConnectionError
from twisted.internet import reactor

from pyfarm.core.enums import OS, WINDOWS, AgentState, INTEGER_TYPES

from pyfarm.agent.logger import getLogger
from pyfarm.agent.logger.twistd import Observer
from pyfarm.agent.config import config
from pyfarm.agent.entrypoints.parser import (
    AgentArgumentParser, ip, port,  uidgid, enum, number, system_identifier)
from pyfarm.agent.entrypoints.utility import (
    start_daemon_posix, get_system_identifier)
from pyfarm.agent.sysinfo import memory, cpu


logger = getLogger("agent.cmd")

config["start"] = time.time()


class AgentEntryPoint(object):
    """Main object for parsing command line options"""
    def __init__(self):
        self.args = None
        self.parser = AgentArgumentParser(
            usage="%(prog)s [status|start|stop]",
            epilog="%(prog)s is a command line client for working with a "
                   "local agent.  You can use it to stop, start, and report "
                   "the general status of a running agent process.")

        # main subparser for start/stop/status/etc
        subparsers = self.parser.add_subparsers(
            help="individual operations %(prog)s can run")
        start = subparsers.add_parser(
            "start", help="starts the agent")
        stop = subparsers.add_parser(
            "stop", help="stops the agent")
        status = subparsers.add_parser(
            "status", help="query the 'running' state of the agent")

        # relate a name and function to each subparser
        start.set_defaults(target_name="start", target_func=self.start)
        stop.set_defaults(target_name="stop", target_func=self.stop)
        status.set_defaults(target_name="status", target_func=self.status)

        # command line flags which configure the agent's network service
        global_network = self.parser.add_argument_group(
            "Agent Network Service",
            description="Main flags which control the network services running "
                        "on the agent.")
        global_network.add_argument(
            "--port", config="agent_api_port", type=port,
            type_kwargs=dict(get_uid=lambda: self.args.uid == 0),
            help="The port number which the agent is either running on or "
                 "will run on when started.  This port is also reported the "
                 "master when an agent starts. [default: %(default)s]")
        global_network.add_argument(
            "--host", config="agent_hostname",
            help="The host to communicate with or hostname to present to the "
                 "master when starting.  Defaults to the fully qualified "
                 "hostname.")
        global_network.add_argument(
            "--agent-api-username", default="agent", config=False,
            help="The username required to access or manipulate the agent "
                 "using REST. [default: %(default)s]")
        global_network.add_argument(
            "--agent-api-password", default="agent", config=False,
            help="The password required to access manipulate the agent "
                 "using REST. [default: %(default)s]")
        global_network.add_argument(
            "--systemid", config="agent_systemid", type=system_identifier,
            default=config["agent_systemid"],
            help="The system identification value.  This is used to help "
                 "identify the system itself to the master when the agent "
                 "connects. [default: %(default)s]")
        global_network.add_argument(
            "--systemid-cache", config="agent_systemid_cache",
            help="The location to cache the value for --systemid. "
                 "[default: %(default)s]")

        # command line flags for the connecting the master apis
        global_apis = self.parser.add_argument_group(
            "Network Resources",
            description="Resources which the agent will be communicating with.")
        global_apis.add_argument(
            "--master", config="master",
            help="This is a convenience flag which will allow you to set the "
                 "hostname for the master.  By default this value will be "
                 "substituted in --master-api")
        global_apis.add_argument(
            "--master-api", config="master_api",
            help="The location where the master's REST api is located. "
                 "[default: %(default)s]")
        global_apis.add_argument(
            "--master-api-version", config="master_api_version",
            help="Sets the version of the master's REST api the agent should"
                 "use [default: %(default)s]")

        # global command line flags which apply to top level
        # process control
        global_process = self.parser.add_argument_group(
            "Process Control",
            description="These settings apply to the parent process of the "
                        "agent and contribute to allowing the process to run "
                        "as other users or remain isolated in an environment. "
                        "They also assist in maintaining the 'running state' "
                        "via a process id file.")
        global_process.add_argument(
            "--pidfile", config="agent_lock_file",
            help="The file to store the process id in. [default: %(default)s]")
        global_process.add_argument(
            "-n", "--no-daemon", default=False, action="store_true",
            config=False,
            help="If provided then do not run the process in the background.")
        global_process.add_argument(
            "--chdir", config="agent_chdir", type=isdir,
            help="The working directory to change the agent into upon launch")
        global_process.add_argument(
            "--uid", type=uidgid, config=False,
            type_kwargs=dict(get_id=getuid, check_id=getpwuid, set_id=setuid),
            help="The user id to run the agent as.  *This setting is "
                 "ignored on Windows.*")
        global_process.add_argument(
            "--gid", type=uidgid, config=False,
            type_kwargs=dict(get_id=getgid, check_id=getgrgid, set_id=setgid),
            help="The group id to run the agent as.  *This setting is "
                 "ignored on Windows.*")
        global_process.add_argument(
            "--pdb-on-unhandled", action="store_true",
            help="When set pdb.set_trace() will be called if an unhandled "
                 "error is caught in the logger")

        # start general group
        start_general_group = start.add_argument_group(
            "General Configuration",
            description="These flags configure parts of the agent related to "
                        "hardware, state, and certain timing and scheduling "
                        "attributes.")
        start_general_group.add_argument(
            "--projects", default=[], nargs="+", config=False,
            help="The project or projects this agent is dedicated to.  By "
                 "default the agent will service any project however specific "
                 "projects may be specified.  For example if you wish this "
                 "agent to service 'Foo Part I' and 'Foo Part II' only just "
                 "specify it as `--projects \"Foo Part I\" \"Foo Part II\"`")
        start_general_group.add_argument(
            "--state", default=AgentState.ONLINE, config=False,
            type=enum, type_kwargs=dict(enum=AgentState),
            help="The current agent state, valid values are "
                 "" + str(list(AgentState)) + ". [default: %(default)s]")
        start_general_group.add_argument(
            "--time-offset", config="agent_time_offset",
            type=int, type_kwargs=dict(min_=0),
            help="If provided then don't talk to the NTP server at all to "
                 "calculate the time offset.  If you know for a fact that this "
                 "host's time is always up to date then setting this to 0 is "
                 "probably a safe bet.")
        start_general_group.add_argument(
            "--ntp-server", config="agent_ntp_server",
            help="The default network time server this agent should query to "
                 "retrieve the real time.  This will be used to help determine "
                 "the agent's clock skew if any.  Setting this value to '' "
                 "will effectively disable this query. [default: %(default)s]")
        start_general_group.add_argument(
            "--ntp-server-version", config="agent_ntp_server_version",
            type=int,
            help="The version of the NTP server in case it's running an older"
                 "or newer version. [default: %(default)s]")
        start_general_group.add_argument(
            "--no-pretty-json", config="agent_pretty_json",
            action="store_false",
            help="If provided do not dump human readable json via the agent's "
                 "REST api")
        start_general_group.add_argument(
            "--shutdown-timeout", config="agent_shutdown_timeout",
            type=int, type_kwargs=dict(min_=0),
            help="How many seconds the agent should spend attempting to inform "
                 "the master that it's shutting down.")
        start_general_group.add_argument(
            "--updates-drop-dir", config="agent_updates_dir",
            help="The directory to drop downloaded updates in. This should be "
            "the same directory pyfarm-supervisor will look for updates in. "
            "[default: %(default)s]")

        # start hardware group
        start_hardware_group = start.add_argument_group(
            "Physical Hardware",
            description="Command line flags which describe the hardware of "
                        "the agent.")
        start_hardware_group.add_argument(
            "--cpus", default=cpu.total_cpus(),
            config="agent_cpus", type=int,
            help="The total amount of cpus installed on the "
                 "system.  Defaults to the number of cpus installed "
                 "on the system.")
        start_hardware_group.add_argument(
            "--ram", default=memory.total_ram(),
            config="agent_ram", type=int,
            help="The total amount of ram installed on the system in "
                 "megabytes.  Defaults to the amount of ram the "
                 "system has installed.")

        # start interval controls
        start_interval_group = start.add_argument_group(
            "Interval Controls",
            description="Controls which dictate when certain internal "
                        "intervals should occur.")
        start_interval_group.add_argument(
            "--ram-check-interval",
            config="agent_ram_check_interval", type=int,
            help="How often ram resources should be checked for changes. "
                 "The amount of memory currently being consumed on the system "
                 "is checked after certain events occur such as a process but "
                 "this flag specifically controls how often we should check "
                 "when no such events are occurring. [default: %(default)s]")
        start_interval_group.add_argument(
            "--ram-max-report-frequency",
            config="agent_ram_max_report_frequency", type=int,
            help="This is a limiter that prevents the agent from reporting "
                 "memory changes to the master more often than a specific "
                 "time interval.  This is done in order to ensure that when "
                 "100s of events fire in a short period of time cause changes "
                 "in ram usage only one or two will be reported to the "
                 "master. [default: %(default)s]")
        start_interval_group.add_argument(
            "--ram-report-delta", config="agent_ram_report_delta", type=int,
            help="Only report a change in ram if the value has changed "
                 "at least this many megabytes. [default: %(default)s]")
        start_interval_group.add_argument(
            "--master-reannounce", config="agent_master_reannounce", type=int,
            help="Controls how often the agent should reannounce itself "
                 "to the master.  The agent may be in contact with the master "
                 "more often than this however during long period of "
                 "inactivity this is how often the agent will 'inform' the "
                 "master the agent is still online.")

        # start logging options
        logging_group = start.add_argument_group(
            "Logging Options",
            description="Settings which control logging of the agent's parent "
                        "process and/or any subprocess it runs.")
        logging_group.add_argument(
            "--log", config="agent_log",
            help="If provided log all output from the agent to this path.  "
                 "This will append to any existing log data.  [default: "
                 "%(default)s]")
        logging_group.add_argument(
            "--capture-process-output", config="jobtype_capture_process_output",
            action="store_true",
            help="If provided then all log output from each process launched "
                 "by the agent will be sent through agent's loggers.")
        logging_group.add_argument(
            "--task-log-dir", config="jobtype_task_logs",
            type=isdir, type_kwargs=dict(create=True),
            help="The directory tasks should log to.")

        # network options for the agent when start is called
        start_network = start.add_argument_group(
            "Network Service",
            description="Controls how the agent is seen or interacted with "
                        "by external services such as the master.")
        start_network.add_argument(
            "--ip-remote", type=ip, config=False,
            help="The remote IPv4 address to report.  In situation where the "
                 "agent is behind a firewall this value will typically be "
                 "different.")

        start_manhole = start.add_argument_group(
            "Manhole Service",
            description="Controls the manhole service which allows a telnet "
                        "connection to be made directly into the agent as "
                        "it's running.")
        start_manhole.add_argument(
            "--enable-manhole", config="agent_manhole",
            action="store_true",
            help="When provided the manhole service will be started once the "
                 "reactor is running.")
        start_manhole.add_argument(
            "--manhole-port", config="agent_manhole_port", type=port,
            type_kwargs=dict(get_uid=lambda: self.args.uid == 0),
            help="The port the manhole service should run on if enabled.")
        start_manhole.add_argument(
            "--manhole-username", config="agent_manhole_username",
            help="The telnet username that's allowed to connect to the "
                 "manhole service running on the agent.")
        start_manhole.add_argument(
            "--manhole-password", config="agent_manhole_password",
            help="The telnet password to use when connecting to the "
                 "manhole service running on the agent.")

        # various options for how the agent will interact with the
        # master server
        start_http_group = start.add_argument_group(
            "HTTP Configuration",
            description="Options for how the agent will interact with the "
                        "master's REST api and how it should run it's own "
                        "REST api.")
        start_http_group.add_argument(
            "--html-templates-reload", config="agent_html_template_reload",
            action="store_true",
            help="If provided then force Jinja2, the html template system, "
                 "to check the file system for changes with every request. "
                 "This flag should not be used in production but is useful "
                 "for development and debugging purposes.")
        start_http_group.add_argument(
            "--static-files", config="agent_static_root", type=isdir,
            help="The default location where the agent's http server should "
                 "find static files to serve.")
        start_http_group.add_argument(
            "--http-retry-delay", config="agent_http_retry_delay", type=number,
            help="If a http request to the master has failed, wait this amount "
                 "of time before trying again")

        jobtype_group = start.add_argument_group("Job Types")
        jobtype_group.add_argument(
            "--jobtype-no-cache", config="jobtype_enable_cache",
            action="store_true",
            help="If provided then do not cache job types, always directly "
                 "retrieve them.  This is beneficial if you're testing the "
                 "agent or a new job type class.")

        # options when stopping the agent
        stop_group = stop.add_argument_group(
            "optional flags",
            description="Flags that control how the agent is stopped")
        stop_group.add_argument(
            "--no-wait", default=False, action="store_true", config=False,
            help="If provided then don't wait on the agent to shut itself "
                 "down.  By default we would want to wait on each task to stop "
                 "so we can catch any errors and then finally wait on the "
                 "agent to shutdown too.  If you're in a hurry or stopping a "
                 "bunch of agents at once then setting this flag will let the "
                 "agent continue to stop itself without waiting for each agent")

    def __call__(self):
        logger.debug("Parsing command line arguments")
        self.args = self.parser.parse_args()

        # No daemon mode must be set with --pdb-on-unhanded, without this
        # you could end up with a blocking agent and no way of knowing
        # about it.
        if not self.args.no_daemon and self.args.pdb_on_unhandled:
            self.parser.error(
                "You cannot set --pdb-on-unhandled without --no-daemon")

        if self.args.pdb_on_unhandled:
            Observer.PDB_ON_UNHANDLED_ERROR = True

            def excepthook(exctype, value, traceback):
                pdb.set_trace()

            sys.excepthook = excepthook

        if not config["master"] and self.args.target_name == "start":
            self.parser.error(
                "--master must be provided (ex. "
                "'pyfarm-agent --master=foobar start')")

        # if we're on windows, produce some warnings about
        # flags which are not supported
        if WINDOWS and self.args.uid:
            logger.warning("--uid is not currently supported on Windows")

        if WINDOWS and self.args.gid:
            logger.warning("--gid is not currently supported on Windows")

        if WINDOWS and self.args.no_daemon:
            logger.warning("--no-daemon is not currently supported on Windows")

        if self.args.target_name == "start":
            # Setup the system identifier
            systemid = get_system_identifier(
                self.args.systemid, config["agent_systemid_cache"])
            config["agent_systemid"] = systemid

            # update configuration with values from the command line
            config_flags = {
                "state": self.args.state,
                "projects": list(set(self.args.projects)),
                "pids": {
                    "parent": os.getpid()}}
            # update configuration with values from the command line

            config.update(config_flags)

        return_code = self.args.target_func()

        # If a specific return code is provided then use it
        # directly in sys.exit
        if isinstance(return_code, INTEGER_TYPES):
            sys.exit(return_code)

    @property
    def agent_api(self):
        return "http://%s:%s/api/v1" % (
            config["agent_hostname"], config["agent_api_port"])

    def start(self):
        url = self.agent_api + "/status"
        try:
            response = requests.get(
                url, headers={"Content-Type": "application/json"})
        except ConnectionError:
            if isfile(config["agent_lock_file"]):
                logger.debug(
                    "Process ID file %s exists", config["agent_lock_file"])

                with open(config["agent_lock_file"], "r") as pidfile:
                    try:
                        pid = int(pidfile.read().strip())
                    except ValueError:
                        logger.warning(
                            "Could not convert pid in %s to an integer.",
                            config["agent_lock_file"])
                    else:
                        try:
                            process = psutil.Process(pid)
                        except psutil.NoSuchProcess:
                            logger.debug(
                                "Process ID in %s is stale.",
                                config["agent_lock_file"])
                            try:
                                os.remove(config["agent_lock_file"])
                            except OSError as e:
                                logger.error(
                                    "Failed to remove PID file %s: %s",
                                    config["agent_lock_file"], e)
                                return 1
                        else:
                            if process.name() == "pyfarm-agent":
                                logger.error(
                                    "Agent is already running, pid %s", pid)
                                return 1
                            else:
                                logger.debug(
                                    "Process %s does not appear to be the "
                                    "agent.", pid)
            else:
                logger.debug(
                    "Process ID file %s does not exist",
                    config["agent_lock_file"])
        else:
            code = "%s %s" % (
                response.status_code, responses[response.status_code])
            pid = response.json()["pids"]["parent"]
            logger.error(
                "Agent at pid %s is already running, got %s from %s.",
                pid, code, url)
            return 1

        logger.info("Starting agent")

        if not isdir(config["jobtype_task_logs"]):
            logger.debug("Creating %s", config["jobtype_task_logs"])
            try:
                os.makedirs(config["jobtype_task_logs"])
            except OSError:
                logger.error("Failed to create %s", config["jobtype_task_logs"])
                return 1

        # create the directory for log
        if not self.args.no_daemon and not isfile(config["agent_log"]):
            try:
                os.makedirs(dirname(config["agent_log"]))
            except OSError:
                # Not an error because it could be created later on
                logger.warning(
                    "failed to create %s" % dirname(config["agent_log"]))

        # so long as fork could be imported and --no-daemon was not set
        # then setup the log files
        if not self.args.no_daemon and fork is not NotImplemented:
            logger.info("sending log output to %s" % config["agent_log"])
            daemon_start_return_code = start_daemon_posix(
                config["agent_log"], config["agent_chdir"],
                self.args.uid, self.args.gid)

            if isinstance(daemon_start_return_code, INTEGER_TYPES):
                return daemon_start_return_code

        elif not self.args.no_daemon and fork is NotImplemented:
            logger.warning(
                "`fork` is not implemented on %s, starting in "
                "foreground" % OS.title())

        # PID file should not exist now.  Either the last agent instance
        # should have removed it or we should hae above.
        if isfile(config["agent_lock_file"]):
            logger.error("PID file should not exist on disk at this point.")
            return 1

        # Create the directory for the pid file if necessary
        pid_dirname = dirname(config["agent_lock_file"])
        if not isdir(pid_dirname):
            try:
                os.makedirs(pid_dirname)
            except OSError:  # pragma: no cover
                logger.error(
                    "Failed to create parent directory for %s",
                    config["agent_lock_file"])
                return 1
            else:
                logger.debug("Created directory %s", pid_dirname)

        # Write the PID file
        try:
            with open(config["agent_lock_file"], "w") as pid:
                pid.write(str(os.getpid()))
        except OSError as e:
            logger.error(
                "Failed to write PID file %s: %s", config["agent_lock_file"], e)
            return 1
        else:
            logger.debug("Wrote PID to %s", config["agent_lock_file"])

        logger.info("pid: %s" % pid)

        if getuid is not NotImplemented:
            logger.info("uid: %s" % getuid())

        if getgid is not NotImplemented:
            logger.info("gid: %s" % getgid())

        from pyfarm.agent.service import Agent

        # Setup the agent, register stop(), then run the agent
        service = Agent()
        signal.signal(signal.SIGINT, service.sigint_handler)
        reactor.callWhenRunning(service.start)
        reactor.run()

    def stop(self):
        logger.info("Stopping the agent")
        url = self.agent_api + "/stop"

        try:
            # TODO: this NEEDS to be an authenticated request
            response = requests.post(
                url,
                data=dumps({"wait": not self.args.no_wait}),
                headers={"Content-Type": "application/json"})

        except Exception as e:
            logger.error("Failed to contact %s: %s", url, e)
            return 1

        if response.status_code == ACCEPTED:
            logger.info("Agent is stopping")
        elif response.status_code == OK:
            logger.info("Agent has stopped")
        else:
            logger.error(
                "Received code %s when attempting to access %s: %s",
                response.status_code, url, response.text)

    def status(self):
        url = self.agent_api + "/status"
        logger.debug("Checking agent status via api using 'GET %s'", url)
        try:
            response = requests.get(
                url, headers={"Content-Type": "application/json"})

        # REST request failed for some reason, try with the pid file
        except Exception as e:
            logger.debug(str(e))
            logger.warning(
                "Failed to communicate with %s's API.  We can only roughly "
                "determine if agent is offline or online.",
                config["agent_hostname"])

            # TODO: use config for pidfile, --pidfile should be an override
            if not isfile(config["agent_lock_file"]):
                logger.debug(
                    "Process ID file %s does not exist",
                    config["agent_lock_file"])
                logger.info("Agent is offline")
                return 1

            else:
                with open(config["agent_lock_file"], "r") as pidfile:
                    try:
                        pid = int(pidfile.read().strip())
                    except ValueError:
                        logger.error(
                            "Could not convert pid in %s to an integer.",
                            config["agent_lock_file"])
                        return 1

                try:
                    process = psutil.Process(pid)
                except psutil.NoSuchProcess:
                    logger.info("Agent is offline.")
                    return 1
                else:
                    if process.name() == "pyfarm-agent":
                        logger.info("Agent is online.")
                    else:
                        logger.warning(
                            "Process %s does not appear to be the agent", pid)
                        logger.info("Agent is offline.")

            return

        data = response.json()
        pid_parent = data["pids"]["parent"]
        pid_child = data["pids"]["child"]

        # Print some general information about the agent
        logger.info("Agent %(agent_hostname)s is %(state)s" % data)
        logger.info("               Uptime: %(uptime)s seconds" % data)
        logger.info(
            "  Last Master Contact: %(last_master_contact)s seconds" % data)
        logger.info("    Parent Process ID: %(pid_parent)s" % locals())
        logger.info("           Process ID: %(pid_child)s" % locals())
        logger.info("          Database ID: %(id)s" % data)
        logger.info("            System ID: %(agent_systemid)s" % data)
        logger.info(
            "      Child Processes: %(child_processes)s "
            "(+%(grandchild_processes)s grandchildren)" % data)
        logger.info("   Memory Consumption: %(consumed_ram)sMB" % data)


agent = AgentEntryPoint()



# No shebang line, this module is meant to be imported
#
# Copyright 2014 Oliver Palmer
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Twisted Logger
--------------

Provides facilities for handling logging events
in a manner which is consistent with Twisted's normal
way of handling log events.
"""

from __future__ import print_function

import pdb
import sys
import logging
from collections import deque
from datetime import datetime
from fnmatch import fnmatch
from logging import DEBUG, INFO, WARNING, CRITICAL, ERROR, FATAL, _levelNames

from twisted.python.failure import Failure
from twisted.python.log import textFromEventDict
from twisted.internet.error import ProcessDone

from pyfarm.core.enums import INTERACTIVE_INTERPRETER

CONFIGURATION = {
    "output": "stdout",
    "datefmt": "%Y-%m-%d %H:%M:%S",
    "format": "%(asctime)s %(levelname)-8s - %(name)-15s - %(message)s",

    # Defines the cutoff level for different loggers.  By default
    # the only defined cutoff is for root ("").  Logger names
    # should be defined using *'s to define matches.
    "levels": [
        ("", DEBUG),
        ("HTTP11ClientProtocol*", INFO)]}

# Only setup colorama if we're not inside
# of an interpreter.
if not INTERACTIVE_INTERPRETER:
    try:
        colorama
    except NameError:
        from colorama import init, Fore, Back, Style
        init()


class Observer(object):
    """
    Main class which provides handling and output
    of log events.
    """
    INSTANCE = None
    PDB_ON_UNHANDLED_ERROR = False

    if not INTERACTIVE_INTERPRETER:
        FORMATS = {
            DEBUG: (Style.DIM, Style.RESET_ALL),
            WARNING: (Fore.YELLOW, Fore.RESET),
            ERROR: (Fore.RED, Fore.RESET),
            CRITICAL: (
                Fore.RED + Style.BRIGHT, Fore.RESET + Style.RESET_ALL),
            FATAL: (
                Fore.RED + Style.BRIGHT, Fore.RESET + Style.RESET_ALL)}

        def add_color(self, text, level):
            try:
                head, tail = self.FORMATS[level]
                return head + text + tail
            except KeyError:
                return text

    else:
        FORMATS = {}
        def add_color(self, text, _):
            return text

    def __init__(self):
        self.backlog = deque()
        self.datefmt = None
        self.format = None
        self.output = None
        self.levels = []

    def configure(self):
        """
        Configures the logger using the :const:`CONFIGURATION` dictionary
        in this module.  This establishes the
        """
        if CONFIGURATION["output"] == "stdout":
            self.output = sys.stdout
        elif CONFIGURATION["output"] == "stderr":
            self.output = sys.stderr
        else:
            raise NotImplementedError(
                "Don't know how to use %r for output" % CONFIGURATION["output"])

        self.datefmt = CONFIGURATION["datefmt"]
        self.format = CONFIGURATION["format"]

    def filter(self, name, level, message):
        """
        Return ``True`` if the given log line should be
        filtered out.
        """
        if message == "Log opened.":
            return True

        for fname, flevel in CONFIGURATION["levels"]:
            if fname in (name, "") or fnmatch(name, fname):
                if flevel > level:
                    return True

        return False

    def emit(self, event):
        """
        Every log message that is produced from from either
        a Python logger or :class:`pyfarm.agent.logger.twistd.Logger`
        will be passed into this method.

        This method will take the logging event, format it, filters
        if necessary and then writes the message to stdout or stderr
        depending on the configuration.
        """
        # Get the message
        text = textFromEventDict(event)
        if text is None:
            return

        # Create a timestamp
        if "time" in event:
            asctime = datetime.fromtimestamp(
                event["time"]).strftime(self.datefmt)
        else:
            # Since we're going to display the time we don't
            # want UTC here.
            asctime = datetime.now().strftime(self.datefmt)

        levelno = event.get("logLevel", DEBUG)
        levelname = _levelNames[levelno]
        name = event.get("system", "twisted")
        message = text % event.get("args", ())

        if name == "-":
            name = "twisted"

        # Log any unhandled error here because it's possible
        # that it won't be handled in a callback
        if event.get("isError") and "failure" in event:
            if isinstance(event["failure"], Failure) \
                    and event["failure"].type is ProcessDone:
                message = "Ignoring previous error.  Error object was " \
                          "'ProcessDone', nothing to do here."
            else:
                logging.exception(message)
                if self.PDB_ON_UNHANDLED_ERROR:
                    pdb.set_trace()

                return

        if self.filter(name, levelno, message):
            return

        print(self.add_color(self.format % locals(), levelno),
              file=self.output)

    def __call__(self, event):
        """
        Handles all logging events which are passed into Twisted's
        logging system.  This will queue logging messages until
        :class:`Observer` has been configured.  Once the observer has
        been configured however logging messages will be passed along
        to :meth:`emit`.
        """
        # Only queue messages if we don't have something
        # to output to.
        if self.output is None:
            self.backlog.append(event)
            return

        while self.backlog:
            try:
                self.emit(self.backlog.pop())
            except IndexError:
                break

        self.emit(event)



# No shebang line, this module is meant to be imported
#
# Copyright 2014 Oliver Palmer
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Utility
=======

Small objects and functions which facilitate operations
on the main entry point class.
"""

import atexit
import os
import sys
from errno import EEXIST, ENOENT
from os.path import dirname

try:
    from os import setuid, setgid, fork
except ImportError:  # pragma: no cover
    setuid = NotImplemented
    setgid = NotImplemented
    fork = NotImplemented


from pyfarm.core.enums import OS, INTEGER_TYPES
from pyfarm.core.utility import convert
from pyfarm.agent.config import config
from pyfarm.agent.logger import getLogger
from pyfarm.agent.sysinfo import system

logger = getLogger("agent.cmd")

SYSTEMID_MAX = 281474976710655


# This is a Linux specific test and will be hard to get due to the nature
# of how the tests are run, for now we're excluding it.
# TODO: figure out a reliable way to test  start_daemon_posix
def start_daemon_posix(log, chdir, uid, gid):  # pragma: no cover
    """
    Runs the agent process via a double fork.  This basically a duplicate
    of Marcechal's original code with some adjustments:

        http://www.jejik.com/articles/2007/02/
        a_simple_unix_linux_daemon_in_python/

    Source files from his post are here:
        http://www.jejik.com/files/examples/daemon.py
        http://www.jejik.com/files/examples/daemon3x.py
    """
    # first fork
    try:
        pid = fork()
        if pid > 0:
            sys.exit(0)
    except OSError as e:
        logger.error(
            "fork 1 failed (errno: %s): %s" % (e.errno, e.strerror))
        return 1

    # decouple from the parent environment
    os.chdir(chdir or "/")
    os.setsid()
    os.umask(0)

    # second fork
    try:
        pid = fork()
        if pid > 0:
            sys.exit(0)
    except OSError as e:
        logger.error(
            "fork 2 failed (errno: %s): %s" % (e.errno, e.strerror))
        return 1

    # flush any pending data before we duplicate
    # the file descriptors
    sys.stdout.flush()
    sys.stderr.flush()

    # open up file descriptors for the new process
    stdin = open(os.devnull, "r")
    logout = open(log, "a+", 0)
    os.dup2(stdin.fileno(), sys.stdin.fileno())
    os.dup2(logout.fileno(), sys.stdout.fileno())
    os.dup2(logout.fileno(), sys.stderr.fileno())

    # if requested, set the user id of this process
    if uid is not None and setuid is not NotImplemented:
        setuid(uid)

    elif uid is not None:
        logger.warning(
            "--uid was requested but `setuid` is not "
            "implemented on %s" % OS.title())

    # if requested, set the group id of this process
    if gid is not None and setgid is not NotImplemented:
        setgid(gid)

    elif gid is not None:
        logger.warning(
            "--gid was requested but `setgid` is not "
            "implemented on %s" % OS.title())


def get_system_identifier(systemid, cache_path=None, write_always=False):
    """
    Generate a system identifier based on the mac addresses
    of this system.  Each mac address is converted to an
    integer then XORed together into an integer value
    no greater than 0xffffffffffff.  This maximum value is
    derived from a mac address which has been maxed out
    ``ff:ff:ff:ff:ff:ff``.

    This value is used to help identify the agent to the
    master more reliably than by other means alone.  In general
    we only create this value once so even if the mac addresses
    should change the return value should not.

    For reference, this function builds a system identifier in
    exactly the same way :func:`uuid.getnode` does.  The main
    differences are that we can handle multiple addresses and
    cache the value between invocations.

    :param int systemid:
        If provided then use this value directly instead of trying
        to generate one.  This is useful if you want to use a
        specific value and provide caching at the same time.

    :param string cache_path:
        If provided then the value will be retrieved from this
        location if it exists.  If the location does not exist
        however this function will generate the value and then
        store it for future use.

    :raises ValueError:
        Raised if ``systemid`` is provided and it's outside the value
        range of input (0 to :const:`SYSTEMID_MAX`)

    :raises TypeError:
        Raised if we receive an unexpected type for one of the inputs
    """
    if not any([systemid is None,
                systemid == "auto",
                isinstance(systemid, INTEGER_TYPES)]):
        raise TypeError("Expected None, 'auto' or an integer for "
                        "`input_systemid`")

    if isinstance(systemid, INTEGER_TYPES) \
            and not 0 < systemid <= SYSTEMID_MAX:
        raise ValueError("input_systemid's range is 0 to %s" % SYSTEMID_MAX)

    if cache_path is None:
        cache_path = config["agent_systemid_cache"]  # pragma: no cover

    write_failed = False
    if systemid == "auto":
        try:
            # Try to read the systemid from the cached file
            with open(cache_path, "r") as cache_file:
                try:
                    systemid = convert.ston(cache_file.read())

                # There's something wrong with the data in the cache
                # file
                except ValueError:  # pragma: no cover
                    logger.warning(
                        "Failed to read cached system identifier from %s, "
                        "this file will be removed.", cache_file.name)

                    try:
                        os.remove(cache_path)
                    except (IOError, OSError) as e:
                        if e.errno != ENOENT:
                            logger.warning(
                                "Failed to remove %s: %s.  Will retry on "
                                "shutdown.", cache_path, e)
                            atexit.register(os.remove, cache_path)

                    # overwrite because there's a problem with the
                    # stored value
                    write_always = True
                else:  # pragma: no cover
                    logger.info("Loaded system identifier %s from %s",
                                systemid, cache_file.name)

        except (OSError, IOError) as e:  # pragma: no cover
            # If the file exists there may be something wrong with it,
            # try to remove it.
            if e.errno != ENOENT:
                logger.error(
                    "System identifier cache file %s exists but it could not "
                    "be read: %s.  The file will be removed.", cache_path, e)
            write_failed = True

    if isinstance(systemid, INTEGER_TYPES) \
            and not 0 < systemid <= SYSTEMID_MAX:  # pragma: no cover
        logger.warning(
            "System identifier from cache is not in range is "
            "0 to 281474976710655.  Cache file will be deleted.")
        try:
            os.remove(cache_path)
        except (IOError, OSError) as e:
            if e.errno != ENOENT:
                logger.warning(
                    "Failed to remove %s: %s.  Will retry on "
                    "shutdown.", cache_path, e)
                atexit.register(os.remove, cache_path)

        write_always = True

    if write_failed or write_always:  # pragma: no cover
        # Create the parent directory if it does not exist
        parent_dir = dirname(cache_path)
        try:
            os.makedirs(parent_dir)
        except (OSError, IOError) as e:
            if e.errno != EEXIST:
                logger.error("Failed to create %r: %s", parent_dir, e)
                raise
        else:
            logger.debug("Created %r", parent_dir)

        # System identifier is either not cache, invalid or
        # none was given.  Generate systemid and cache it.
        systemid = system.system_identifier()
        try:
            with open(cache_path, "w") as cache_file:
                cache_file.write(str(systemid))
        except (OSError, IOError) as e:
            logger.warning(
                "Failed to cache system identifier to %s: %s", systemid, e)
        else:
            logger.info(
                "Cached system identifier %s to %s",
                systemid, cache_file.name)

        if isinstance(systemid, INTEGER_TYPES) \
                and not 0 < systemid <= SYSTEMID_MAX:
            raise ValueError("systemid's range is 0 to %s" % SYSTEMID_MAX)

        return systemid

    logger.debug("Custom system identifier will not be cached.")
    return systemid



# No shebang line, this module is meant to be imported
#
# Copyright 2013 Oliver Palmer
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Resource
--------

Base resources which can be used to build top leve
documents, pages, or other types of data for the web.
"""

from json import loads

try:
    from httplib import (
        responses, NOT_FOUND, BAD_REQUEST, UNSUPPORTED_MEDIA_TYPE,
        METHOD_NOT_ALLOWED)
except ImportError:  # pragma: no cover
    from http.client import (
        responses, NOT_FOUND, BAD_REQUEST, UNSUPPORTED_MEDIA_TYPE,
        METHOD_NOT_ALLOWED)

try:
    from itertools import ifilter as filter_
except ImportError:  # pragma: no cover
    filter_ = filter

try:
    from itertools import imap as map_
except ImportError:  # pragma: no cover
    map_ = map

from twisted.web.server import NOT_DONE_YET
from twisted.web.resource import Resource as _Resource
from voluptuous import Invalid, Schema

from pyfarm.core.enums import STRING_TYPES
from pyfarm.agent.http.core import template
from pyfarm.agent.logger import getLogger

logger = getLogger("agent.http.resource")


class Resource(_Resource):
    """
    Basic subclass of :class:`._Resource` for passing requests to
    specific methods.  Unlike :class:`._Resource` however this will
    will also handle:

        * rewriting of request objects
        * templating
        * content type discovery and validation
        * unpacking of request data
        * rerouting of request to specific internal methods
    """
    TEMPLATE = NotImplemented
    CONTENT_TYPES = set(["text/html", "application/json"])
    LOAD_DATA_FOR_METHODS = set(["POST", "PUT"])

    # Used by APIResource
    SCHEMAS = {}

    def __init__(self):
        _Resource.__init__(self)
        assert isinstance(self.CONTENT_TYPES, set)

    @property
    def template(self):
        """
        Loads the template provided but the partial path in ``TEMPLATE`` on
        the class.
        """
        if self.TEMPLATE is NotImplemented:
            raise NotImplementedError("You must set `TEMPLATE` first")
        return template.load(self.TEMPLATE)

    @property
    def methods(self):
        """A set containing all the methods this resource implements."""
        methods = set()
        for method in ("get", "post", "put", "delete", "head"):
            attribute_count = 0
            for attribute_name in (method, "render_%s" % method.upper()):
                attribute = getattr(self, attribute_name, None)
                if attribute is not None:
                    attribute_count += 1
                    methods.add(method)

            if attribute_count == 2:
                raise ValueError(
                    "%s has both `%s` and `%s` methods" % (
                        self.__class__.__name__,
                        method, "render_%s" % method.upper()))

        return methods

    def content_types(self, request, default=None):
        """Returns the content type(s) present in the request"""
        headers = request.requestHeaders.getRawHeaders("content-type")
        if isinstance(default, STRING_TYPES):
            default = [default]
        elif default is None:
            default = []
        return set(headers) if headers is not None else set(default)

    def putChild(self, path, child):
        """
        Overrides the builtin putChild() so we can return the results for
        each call and use them externally
        """
        _Resource.putChild(self, path, child)
        return child

    def error(self, request, code, message):
        """
        Writes the proper out an error response message depending on the
        content type in the request
        """
        content_types = self.content_types(request, default="text/html")
        logger.error(message)

        if "text/html" in content_types:
            request.setResponseCode(code)
            html_error = template.load("error.html")
            deferred = html_error.render(
                code=code, code_msg=responses[code], message=message)
            deferred.addCallback(request.write).addCallback(
                lambda _: request.finish())

        elif "application/json" in content_types:
            request.setResponseCode(code)
            request.write({"error": message})
            request.finish()

        else:
            request.setResponseCode(UNSUPPORTED_MEDIA_TYPE)
            request.write(
                {"error": "Can only handle text/html or application/json here"})
            request.finish()

    def render(self, request):
        # make sure that the requested content type is supported
        content_type = self.content_types(request, default=["text/html",
                                                            "application/json"])
        if not self.CONTENT_TYPES & content_type:
            self.error(
                request, UNSUPPORTED_MEDIA_TYPE,
                "%s is not a support content type for this url" % content_type)
            return NOT_DONE_YET

        # Try to find the method this web request is making by first trying
        # our usual convention then the 'standard' convention.
        request_methods = (request.method.lower(), "render_%s" % request.method)
        for method_name in request_methods:
            if hasattr(self, method_name):
                kwargs = {"request": request}

                # Unpack the incoming data for the request
                if "application/json" in content_type \
                        and request.method in self.LOAD_DATA_FOR_METHODS:
                    request_content = request.content.read()

                    # Check to see if we have any incoming data at all
                    if not request_content.strip():
                        self.error(request, BAD_REQUEST, "No data provided")
                        return NOT_DONE_YET

                    # Either load the data or handle the error, don't call
                    # the method unless we're successful.
                    try:
                        data = loads(request_content)

                    except ValueError as e:
                        self.error(
                            request, BAD_REQUEST,
                            "Failed to decode json data: %r" % e)
                        return NOT_DONE_YET

                    # We have data, check to see if we have a schema
                    # and if we do does it validate.
                    schema = self.SCHEMAS.get(request.method)
                    if isinstance(schema, Schema):
                        try:
                            schema(data)
                        except Invalid as e:
                            self.error(
                                request, BAD_REQUEST,
                                "Failed to validate the request data "
                                "against the schema: %s" % e)
                            return NOT_DONE_YET

                    kwargs.update(data=data)

                return getattr(self, method_name)(**kwargs)

        # If we could not find function to call for the given method
        # produce an error.
        else:
            supported_methods = self.methods
            message = "%r only supports the %s method(s)" % (
                request.uri,
                ", ".join(list(map_(str.upper, supported_methods))))

            self.error(request, METHOD_NOT_ALLOWED, message)
            return NOT_DONE_YET



# No shebang line, this module is meant to be imported
#
# Copyright 2014 Oliver Palmer
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Manhole
=======

Provides a way to access the internals of the agent via
the telnet protocol.
"""

# NOTE: This module is based on buildbot's manhole implementation
#   https://github.com/buildbot/buildbot/blob/
#   1d35e1cb25dfcdd3d30263bdfce8910a56874e60/master/buildbot/manhole.py

from __future__ import print_function

from pprint import pprint
from inspect import ismethod

from twisted.conch.insults.insults import ServerProtocol
from twisted.conch.manhole import ColoredManhole
from twisted.conch.telnet import (
    ITelnetProtocol, TelnetBootstrapProtocol, AuthenticatingTelnetProtocol,
    TelnetTransport)
from twisted.cred.checkers import InMemoryUsernamePasswordDatabaseDontUse
from twisted.cred.portal import IRealm, Portal
from twisted.internet.protocol import ServerFactory
from zope.interface import implements

from pyfarm.core.enums import STRING_TYPES, INTEGER_TYPES, NOTSET
from pyfarm.agent.logger import getLogger

ITERABLE_TYPES = (list, tuple, dict)

logger = getLogger("agent.manhole")


class LoggingManhole(ColoredManhole):
    """
    A slightly modified implementation of :class:`ColoredManhole`
    which logs information to the logger so we can track activity in
    the agent's log.
    """
    def connectionMade(self):  # pragma: no cover
        peer = self.terminal.transport.getPeer()
        logger.info("Connection made from %s@%s", peer.host, peer.port)
        super(LoggingManhole, self).connectionMade()

    def connectionLost(self, reason):  # pragma: no cover
        peer = self.terminal.transport.getPeer()
        logger.info("Connection lost from %s@%s", peer.host, peer.port)
        super(LoggingManhole, self).connectionLost(reason)

    def lineReceived(self, line):
        peer = self.terminal.transport.getPeer()
        logger.info("%s@%s - %s", peer.host, peer.port, line)

        if line.strip() in ("exit", "exit()", "quit", "quit()", "\q"):
            self.handle_QUIT()
        else:  # pragma: no cover
            super(LoggingManhole, self).lineReceived(line)


class TransportProtocolFactory(object):
    """
    Glues together a portal along with the :class:`TelnetTransport` and
    :class:`AuthenticatingTelnetProtocol` objects.  This class is instanced
    onto the ``protocol`` attribute of the :class:`ServerFactory` class
    in :func:`build_manhole`.
    """
    def __init__(self, portal):
        self.portal = portal

    def __call__(self):
        return TelnetTransport(AuthenticatingTelnetProtocol, self.portal)


class TelnetRealm(object):
    """Wraps together :class:`ITelnetProtocol`,
    :class:`TelnetBootstrapProtocol`, :class:`ServerProtocol` and
    :class:`ColoredManhole` in :meth:`requestAvatar` which will provide
    the interface to the manhole.
    """
    implements(IRealm)
    NAMESPACE = None

    def requestAvatar(self, _, *interfaces):
        if ITelnetProtocol in interfaces:
            return (
                ITelnetProtocol,
                TelnetBootstrapProtocol(
                    ServerProtocol, LoggingManhole,
                    self.NAMESPACE),
                lambda: None)
        raise NotImplementedError


def show(x=NOTSET):
    """Display the data attributes of an object in a readable format"""
    if x is NOTSET:
        print("objects: %s" % TelnetRealm.NAMESPACE.keys())
        return

    print("data attributes of %r" % (x,))

    names = dir(x)
    maxlen = max([0] + [len(n) for n in names])
    for k in names:
        v = getattr(x, k)

        if ismethod(v) or k[:2] == "__" and k[-2:] == "__":
            continue

        if isinstance(v, STRING_TYPES):
            if len(v) > 80 - maxlen - 5:
                v = repr(v[:80 - maxlen - 5]) + "..."

        elif isinstance(v, INTEGER_TYPES) or v is None:
            v = str(v)

        elif isinstance(v, ITERABLE_TYPES):
            v = "%s (%d elements)" % (v, len(v))

        print("%*s : %s" % (maxlen, k, v))
    return x


def manhole_factory(namespace, username, password):
    """
    Produces a factory object which can be used to listen for telnet
    connections to the manhole.
    """
    assert isinstance(namespace, dict)
    assert isinstance(username, STRING_TYPES)
    assert isinstance(password, STRING_TYPES)
    assert TelnetRealm.NAMESPACE is None, "namespace already set"

    # TODO: we should try to use the system to authorize users instead
    checker = InMemoryUsernamePasswordDatabaseDontUse()
    checker.addUser(username, password)

    # Setup the namespace
    namespace = namespace.copy()
    namespace.setdefault("pp", pprint)
    namespace.setdefault("show", show)

    realm = TelnetRealm()
    TelnetRealm.NAMESPACE = namespace
    portal = Portal(realm, [checker])
    factory = ServerFactory()
    factory.protocol = TransportProtocolFactory(portal)
    return factory
